"""Project: Eskapade - A python-based package for data analysis.

Macro: esk206_merge_pandas_dfs

Created: 2017/02/23

Description:
    Illustrate link that calls basic merge() of pandas dataframes
    For more information see pandas documentation:

    http://pandas.pydata.org/pandas-docs/stable/merging.html

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from pandas import DataFrame

from eskapade import ConfigObject, DataStore, Chain
from eskapade import core_ops, analysis
from eskapade import process_manager
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk206_merge_pandas_dfs')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk206_merge_pandas_dfs'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# generate dummy dataframes and add to datastore
# these dataframes are concatenated below, during link execution.

# NB: realize that, normally, such dataframes are read or generated
# on the fly during link execution.

ds = process_manager.service(DataStore)

ds['left'] = DataFrame({'key': ['K0', 'K1', 'K2', 'K3'],
                        'A': ['A0', 'A1', 'A2', 'A3'],
                        'B': ['B0', 'B1', 'B2', 'B3']})

ds['right'] = DataFrame({'key': ['K0', 'K1', 'K2', 'K3'],
                         'C': ['C0', 'C1', 'C2', 'C3'],
                         'D': ['D0', 'D1', 'D2', 'D3']})

#########################################################################################
# --- below we merge the two dataframes found in the datastore

data_prep = Chain('DataPrep')

# inner-join the two dataframes with each other on 'key' during link execution
link = analysis.DfMerger(input_collection1='left',
                         input_collection2='right',
                         output_collection='outgoing',
                         how='inner',
                         on='key')
# Any other kwargs given to DfMerger are passed on the the
# pandas merge() function.
link.logger.log_level = LogLevel.DEBUG
data_prep.add(link)

link = core_ops.DsObjectDeleter()
link.deletion_keys = ['left', 'right']
data_prep.add(link)

link = core_ops.PrintDs()
link.keys = ['n_outgoing', 'outgoing']
data_prep.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk206_merge_pandas_dfs')



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017-08-08

Description:
    Collection of eskapade entry points

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import _bootstrap as bootstrap
from eskapade.logger import LogLevel, Logger, global_log_publisher, ConsoleHandler, ConsoleErrHandler

publisher = global_log_publisher
publisher.log_level = LogLevel.INFO
publisher.add_handler(ConsoleHandler())
publisher.add_handler(ConsoleErrHandler())

logger = Logger(__name__)

DEFAULT_LINK_NAME = 'ExampleLink'
DEFAULT_MACRO_NAME = 'macro'
DEFAULT_NOTEBOOK_NAME = 'notebook'


def eskapade_ignite():
    """Log info message."""
    logger.info('Boom!')


def eskapade_run():
    """Run Eskapade.

    Top-level entry point for an Eskapade run started from the
    command line.  Arguments specified by the user are parsed and
    converted to settings in the configuration object.  Optionally, an
    interactive Python session is started when the run is finished.
    """
    from eskapade import process_manager, ConfigObject, DataStore
    from eskapade.core import execution
    from eskapade.core.run_utils import create_arg_parser

    # create parser for command-line arguments
    parser = create_arg_parser()
    user_args = parser.parse_args()

    # create config object for settings
    if not user_args.unpickle_config:
        # create new config
        settings = ConfigObject()
    else:
        # read previously persisted settings if pickled file is specified
        conf_path = user_args.config_files.pop(0)
        settings = ConfigObject.import_from_file(conf_path)
    del user_args.unpickle_config

    # set configuration macros
    settings.add_macros(user_args.config_files)

    # set user options
    settings.set_user_opts(user_args)

    try:
        # run Eskapade
        execution.eskapade_run(settings)
    except Exception as exc:
        logger.error('{exc}', exc=exc)
        raise

    # start interpreter if requested (--interactive on command line)
    if settings.get('interactive'):
        # set Pandas display options
        import pandas as pd
        pd.set_option('display.width', 120)
        pd.set_option('display.max_columns', 50)

        # make datastore and config available in interactive session
        ds = process_manager.service(DataStore)
        settings = process_manager.service(ConfigObject)

        # start interactive session
        from code import InteractiveConsole
        cons = InteractiveConsole(locals())
        cons.interact("\nContinuing interactive session ... press Ctrl+d to exit.\n")


def eskapade_trial():
    """Run Eskapade tests.

    We will keep this here until we've completed switch to pytest or nose and tox.
    We could also keep it, but I don't like the fact that packages etc. are
    hard coded. Gotta come up with
    a better solution.
    """
    import sys
    import pytest

    # ['--pylint'] +
    # -r xs shows extra info on skips and xfails.
    default_options = ['-rxs']
    args = sys.argv[1:] + default_options
    sys.exit(pytest.main(args))


def _add_project_dir_argument(parser, arg_name, dir_type):
    """Add a project directory argument to a parser.

    :param parser: argparse
    :param arg_name: argument's name
    :param dir_type: type of the directory to be created
    """
    import os

    parser.add_argument('--{arg_name}'.format(arg_name=arg_name),
                        nargs='?',
                        default=os.getcwd(),
                        help='The analysis project {dir_type} directory. Default is: {default}.'
                        .format(dir_type=dir_type, default=os.getcwd()), )


def eskapade_generate_link():
    """Generate Eskapade link.

    By default does not create init file.
    """
    import argparse

    parser = argparse.ArgumentParser('eskapade_generate_link',
                                     description='Generate Eskapade link.')
    parser.add_argument('name',
                        help='The name of the link to generate.', )
    _add_project_dir_argument(parser, 'dir', 'links')
    parser.add_argument('--create_init',
                        nargs='?',
                        default=False,
                        type=bool,
                        help='Whether to create __init__.py file or no. Default is: False.', )
    args = parser.parse_args()

    if args.name == 'Link':
        raise AttributeError('Link is reserved by Eskapade. Please, choose different name for the link.')

    path = bootstrap.get_absolute_path(args.dir)

    bootstrap.create_dir(path)
    bootstrap.generate_link(path, args.name, args.create_init)


def eskapade_generate_macro():
    """Generate Eskapade macro."""
    import argparse

    parser = argparse.ArgumentParser('eskapade_generate_macro',
                                     description='Generate Eskapade macro.')
    parser.add_argument('name',
                        help='The name of the macro to generate.', )
    _add_project_dir_argument(parser, 'dir', 'macros')
    args = parser.parse_args()

    path = bootstrap.get_absolute_path(args.dir)

    bootstrap.create_dir(path)
    bootstrap.generate_macro(macro_dir=path, macro_name=args.name, is_create_init=False)


def eskapade_generate_notebook():
    """Generate Eskapade notebook."""
    import argparse

    parser = argparse.ArgumentParser('eskapade_generate_notebook',
                                     description='Generate Eskapade notebook.')
    parser.add_argument('name',
                        help='The name of the notebook to generate.', )
    _add_project_dir_argument(parser, 'dir', 'notebooks')
    args = parser.parse_args()

    path = bootstrap.get_absolute_path(args.dir)

    bootstrap.create_dir(path)
    bootstrap.generate_notebook(notebook_dir=path, notebook_name=args.name)


def eskapade_bootstrap():
    """Generate Eskapade project structure."""
    import argparse

    parser = argparse.ArgumentParser('eskapade_bootstrap',
                                     description='Generate Eskapade project structure.',
                                     epilog='Please note, existing files with the same names will be rewritten.')
    parser.add_argument('package_name',
                        help='The name of the package to generate.', )
    _add_project_dir_argument(parser, 'project_root_dir', 'root')
    parser.add_argument('--macro_name', '-m',
                        nargs='?',
                        help='The name of the macro to generate. Default is: macro.',
                        default=DEFAULT_MACRO_NAME, )
    parser.add_argument('--link_name', '-l',
                        nargs='?',
                        help='The name of the link to generate. Default is: ExampleLink.',
                        default=DEFAULT_LINK_NAME, )
    parser.add_argument('--notebook_name', '-n',
                        nargs='?',
                        help='The name of the notebook to generate. Default is: notebook.',
                        default=DEFAULT_NOTEBOOK_NAME, )
    args = parser.parse_args()

    if args.link_name == 'Link':
        raise AttributeError('Link is reserved by Eskapade. Please, choose different name for the link.')
    if args.package_name.lower() == 'eskapade':
        raise AttributeError('eskapade is reserved by Eskapade. Please, choose different name for the project.')
    if not args.project_root_dir:
        raise AttributeError('Expected the value after --project_root_dir.')

    package_dir = bootstrap.get_absolute_path(args.project_root_dir) + '/' + args.package_name
    link_dir = package_dir + '/links'
    marco_path = package_dir + '/' + args.macro_name + '.py'

    # create the directories
    bootstrap.create_dir(link_dir)
    bootstrap.generate_link(link_dir=link_dir, link_name=args.link_name, is_create_init=True)
    bootstrap.generate_macro(macro_dir=package_dir, macro_name=args.macro_name,
                             link_module=args.package_name, link_name=args.link_name, is_create_init=True)
    bootstrap.generate_notebook(notebook_dir=package_dir, notebook_name=args.notebook_name, macro_path=marco_path)
    bootstrap.generate_configs(root_dir=package_dir)
    bootstrap.generate_setup(root_dir=package_dir, package_name=args.package_name)



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk205_concatenate_pandas_dfs

Created: 2017/02/23

Description:
    Illustrates link that calls basic concat() of pandas dataframes
    See for more information pandas documentation:

    http://pandas.pydata.org/pandas-docs/stable/merging.html

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from pandas import DataFrame

from eskapade import ConfigObject, DataStore, Chain
from eskapade import core_ops, analysis
from eskapade import process_manager
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk205_concatenate_pandas_dfs.')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk205_concatenate_pandas_dfs'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# generate dummy dataframes and add to datastore
# these dataframes are concatenated below, during link execution.

# NB: realize that, normally, such dataframes are read or generated
# on the fly during link execution.

ds = process_manager.service(DataStore)

ds['df1'] = DataFrame({'A': ['A0', 'A1', 'A2', 'A3'],
                       'B': ['B0', 'B1', 'B2', 'B3'],
                       'C': ['C0', 'C1', 'C2', 'C3'],
                       'D': ['D0', 'D1', 'D2', 'D3']})

ds['df2'] = DataFrame({'A': ['A4', 'A5', 'A6', 'A7'],
                       'B': ['B4', 'B5', 'B6', 'B7'],
                       'C': ['C4', 'C5', 'C6', 'C7'],
                       'D': ['D4', 'D5', 'D6', 'D7']})

ds['df3'] = DataFrame({'A': ['A8', 'A9', 'A10', 'A11'],
                       'B': ['B8', 'B9', 'B10', 'B11'],
                       'C': ['C8', 'C9', 'C10', 'C11'],
                       'D': ['D8', 'D9', 'D10', 'D11']})

#########################################################################################
# --- below we concatenate the dataframes found in the datastore

data_prep = Chain('DataPrep')

# concatenate the three dataframes below each other during link execution
link = analysis.DfConcatenator(read_keys=['df1', 'df2', 'df3'],
                               store_key='outgoing')

# Any other kwargs given to DfConcatenator are passed on the the
# pandas concat() function.
link.logger.log_level = LogLevel.DEBUG
data_prep.add(link)

link = core_ops.DsObjectDeleter()
link.deletion_keys = ['df1', 'df2', 'df3']
data_prep.add(link)

link = core_ops.PrintDs()
link.keys = ['n_outgoing', 'outgoing']
data_prep.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk205_concatenate_pandas_dfs.')



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk501_fix_pandas_dataframe

Created: 2017/04/26

Description:
    Macro illustrates how to call FixPandasDataFrame link that gives columns
    consistent names and datatypes.
    Default settings perform the following clean-up steps on an
    input dataframe:

    - Fix all column names. Eg. remove punctuation and strange characters,
    and convert spaces to underscores.
    - Check for various possible nans in the dataset, then make all nans
    consistent by turning them into numpy.nan (= float)
    - Per column, assess dynamically the most consistent datatype (ignoring
    all nans in that column). Eg. bool, int, float, datetime64, string.
    - Per column, make the data types of all rows consistent, by using the
    identified (or imposed) data type (by default ignoring all nans)

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import tempfile

from eskapade import ConfigObject, Chain
from eskapade import core_ops, analysis, data_quality
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk501_fix_pandas_dataframe')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk501_fix_pandas_dataframe'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# dummy dataframe filled with inconsistent data types per column

tmp = b"""A,B,C,D,E,F,G,H
True,foo,1.0,1,1,1,a,a
False,bar,2.0,2,2,2.5,b,b
nan,3,bal,3,bla,bar,c,1
,nan,NaN,NaN,nan,nan,d,2
,,,,,,,3
1,2,,,,,,,6
"""

f = tempfile.NamedTemporaryFile(delete=False)
f.write(tmp)
f.close()
# file is not immediately deleted because we used delete=False
# used below with f.name

#########################################################################################
# --- now set up the chains and links based on configuration flags

ch = Chain('DataPrep')

# --- 0. pandas read_csv has multiple settings to help reading in of buggy csv's.
#     o The option error_bad_lines=False skips lines with too few or too many values
#     o The option encoding='latin1' interprets most non-standard characters
read_data = analysis.ReadToDf(key='vrh',
                              reader='csv',
                              path=f.name,
                              error_bad_lines=False,
                              encoding='latin1')
ch.add(read_data)

# --- 1. standard setting:
#     o convert all nans to np.nan (= float)
#     o convert all rows in a column to most occuring datatype in that column
fixer = data_quality.FixPandasDataFrame(name='fixer1')
fixer.read_key = 'vrh'
fixer.store_key = 'vrh_fix1'
ch.add(fixer)

# --- 2. force certain columns to specified datatype
fixer = data_quality.FixPandasDataFrame(name='fixer2')
fixer.read_key = 'vrh'
fixer.store_key = 'vrh_fix2'
fixer.var_dtype = {'B': int, 'C': str}
ch.add(fixer)

# --- 3. convert all nans to data type consistent with rest of column
fixer = data_quality.FixPandasDataFrame(name='fixer3')
fixer.read_key = 'vrh'
fixer.store_key = 'vrh_fix3'
fixer.convert_inconsistent_nans = True
# set a specific nan (GREPME) for a given column (G)
fixer.var_nan = {'G': 'GREPME'}
ch.add(fixer)

# --- 4. compare results
pds = core_ops.PrintDs(name='pds2')
pds.keys = ['vrh', 'vrh_fix1', 'vrh_fix2', 'vrh_fix3']
ch.add(pds)

# --- 5. write out fixed dataframe - turned off in this example
writedata = analysis.WriteFromDf(name='writer',
                                 key='vrh_fix1',
                                 path='tmp.csv',
                                 writer='csv')
# ch.add(writedata)

#########################################################################################

logger.debug('Done parsing configuration file esk501_fix_pandas_dataframe')



"""Project: Eskapade - A python-based package for data analysis.

Created: 2018/06/23

Description:
    Convert Pearson correlation value into a chi2 value of a contingency test 
    matrix of a bivariate gaussion, and vice-versa. 
    Calculation uses scipy's mvn library.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import numpy as np
from scipy.stats import mvn
from scipy import optimize

def _mvn_un(rho,lower,upper):
    '''Perform integral of bivariate normal gauss with correlation

    Integral is performed using scipy's mvn library.
    
    :returns float: integral value
    '''
    mu = np.array([0., 0.])
    S = np.array([[1.,rho],[rho,1.0]])
    p,i = mvn.mvnun(lower,upper,mu,S)
    return p

def _mvn_array(rho, sx, sy):
    '''Array of integrals over bivariate normal gauss with correlation

    Integrals are performed using scipy's mvn library.
    
    :returns list: list of integral values
    '''
    corr = []
    for i in range(len(sx)-1):
        for j in range(len(sy)-1):
            lower = [sx[i],sy[j]]
            upper = [sx[i+1],sy[j+1]]
            p = _mvn_un(rho,lower,upper)
            corr.append(p)
    return corr

def chi2_from_rho(rho, n, subtract_from_chi2=0, corr0=None, sx=None, sy=None, nx=-1, ny=-1):
    '''Calculate chi2-value of bivariate gauss having correlation value rho
    
    Calculate no-noise chi2 value of bivar gauss with correlation rho,
    with respect to bivariate gauss without any correlation.
    
    :returns float: chi2 value    
    '''
    assert nx>1 or sx is not None, 'number of bins along x-axis is unknown'
    assert ny>1 or sy is not None, 'number of bins along y-axis is unknown'
    if sx is None:
        sx = np.linspace(-5,5,nx+1)
    if sy is None:
        sy = np.linspace(-5,5,ny+1)
    if corr0 is None:
        corr0 = _mvn_array(0, sx, sy)
    corrr = _mvn_array(rho, sx, sy)
    chi2 = n * sum([((cr-c0)*(cr-c0)) / c0 for c0,cr in zip(corr0,corrr)])
    return chi2 - subtract_from_chi2

def rho_from_chi2(chi2, n, nx, ny, sx=None, sy=None):
    '''correlation coefficient of bivariate gaussian derived from chi2-value
    
    Chi2-value gets converted into correlation coefficient of bivariate gauss
    with correlation value rho, assuming giving binning and number of records. 
    Correlation coefficient value is between 0 and 1.

    Bivariate gaussian's range is set to [-5,5] by construction.

    :returns float: correlation coefficient
    '''
    assert nx>1 or sx is not None, 'number of bins along x-axis is unknown'
    assert ny>1 or sy is not None, 'number of bins along y-axis is unknown'
    if sx is None:
        sx = np.linspace(-5,5,nx+1)
    if sy is None:
        sy = np.linspace(-5,5,ny+1)
    corr0 = _mvn_array(0, sx, sy)
    rho = optimize.brentq(chi2_from_rho, 0, 1, args=(n, chi2, corr0, sx, sy))
    return rho



"""Project: Eskapade - A python-based package for data analysis.

Class: IPythonEmbed

Created: 2017/02/26

Description:
    Link that starts up a python console during execution for debugging.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import Link
from eskapade import StatusCode


class IPythonEmbed(Link):
    """Link to start up a python console.

    Start up a python console by simply adding this link at any location in a chain.
    Note: not an ipython console, but regular python console.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'IPythonEmbed'))

        # check residual kwargs. exit if any present.
        self.check_extra_kwargs(kwargs)

    def execute(self):
        """Execute the link."""
        # this function calls the python session
        # in this session ds, settings, and process_manager are available
        from code import InteractiveConsole
        cons = InteractiveConsole(locals())
        cons.interact("\nStarting interactive session ... press Ctrl+d to exit.\n")

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk202_writedata

Created: 2017/02/20

Description:
    Macro to illustrate writing pandas dataframes to file.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain, resources
from eskapade import analysis
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk202_writedata')

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk202_writedata'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

settings['do_readdata'] = True
settings['do_writedata'] = True

#########################################################################################
# --- Set path of data
data_path = resources.fixture('dummy.csv')

#########################################################################################
# --- now set up the chains and links based on configuration flags

# --- readdata with default settings reads all three input files simultaneously.
#     all extra key word arguments are passed on to pandas reader.
if settings['do_readdata']:
    read = Chain('ReadData')
    # --- readdata keeps on opening the next file in the file list.
    #     all kwargs are passed on to pandas file reader.
    read_data = analysis.ReadToDf(name='reader', key='test', sep='|', reader='csv', path=[data_path] * 3)
    read.add(read_data)

if settings['do_writedata']:
    write = Chain('WriteData')
    # --- writedata needs a specified output format ('writer' argument).
    #     if this is not set, try to determine this from the extension from the filename.
    #     'key' is picked up from the datastore. 'path' is the output filename.
    #     all other kwargs are passed on to pandas file writer.
    write_data = analysis.WriteFromDf(name='writer', key='test', path='tmp3.csv', writer='csv')
    write.add(write_data)

#########################################################################################

logger.debug('Done parsing configuration file esk202_writedata')



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk101_helloworld

Created: 2017/02/20

Description:
    Macro to say hello to the world with Eskapade!

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import core_ops
from eskapade import process_manager
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk101_helloworld')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk101_helloworld'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

#     E.g. define flags turn on or off certain chains with links.
#     by default all set to false, unless already configured in
#     configobject or vars()

settings['do_hello'] = True
settings['n_repeat'] = 2

#########################################################################################
# --- now set up the chains and links based on configuration flags

if settings['do_hello']:
    hello = Chain(name='Hello')
    link = core_ops.HelloWorld(name='HelloWorld')
    link.logger.log_level = LogLevel.DEBUG
    link.repeat = settings['n_repeat']
    hello.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk101_helloworld')



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017-09-20

Description:
    Helper functions for eskapade_bootstrap

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import datetime
import os
import sys

from eskapade.logger import Logger

logger = Logger(__name__)


def get_absolute_path(path):
    """Get an absolute path.

    First expands ~ if present. Second take care of any . or ..

    :param path: path
    :returns: the absolute path
    """
    return os.path.abspath(os.path.expanduser(path))


def create_dir(path):
    """Create a leaf directory and all intermediate ones.

    Exit with a status code other than 0 if there is an error.

    :param path: an absolute path to the directory
    """
    try:
        logger.info('Creating the directory {dir!s}.', dir=path)
        os.makedirs(path, exist_ok=True)
    except PermissionError as exc:
        logger.error('Failed to create the directory {dir!s}! error={err!s}.', dir=path, err=exc.strerror)
        sys.exit(exc.errno)


def create_file(path, file_name, content=''):
    """Create a file in a given directory.

    Exit with a status code other than 0 if there is an error.

    :param path: an absolute path to the directory
    :param file_name: file name
    :param content: file's content
    :returns: path to created file
    """
    try:
        logger.info('Creating {file_name} in the directory {dir!s}.', file_name=file_name, dir=path)
        fp = open(path + '/' + file_name, 'w')
    except PermissionError as exc:
        logger.error('Failed to create {file_name} in the directory {dir!s}! error={err!s}.',
                     file_name=file_name, dir=path, err=exc.strerror)
        sys.exit(exc.errno)
    else:
        with fp:
            fp.write(content)


def generate_link(link_dir, link_name, is_create_init=False):
    """Generate Eskapade link.

    :param link_dir: absolute path to a directory where the link will be generated
    :param link_name: name of the link to generate
    :param is_create_init: whether to create __init__.py file or no
    """
    # Do not modify the indentation of template!
    template = """\"\"\"Project: Eskapade - A python-based package for data analysis.

Class: {link_name!s}

Created: {date_generated!s}

Description:
    Algorithm to ...(fill in one-liner here)

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
\"\"\"

from eskapade import process_manager, ConfigObject, DataStore, Link, StatusCode


class {link_name!s}(Link):

    \"\"\"Defines the content of link.\"\"\"

    def __init__(self, **kwargs):
        \"\"\"Initialize an instance.

        :param str name: name of link
        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        \"\"\"
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', '{link_name!s}'))

        # Process and register keyword arguments. If the arguments are not given, all arguments are popped from
        # kwargs and added as attributes of the link. Otherwise, only the provided arguments are processed.
        self._process_kwargs(kwargs, read_key=None, store_key=None)

        # check residual kwargs; exit if any present
        self.check_extra_kwargs(kwargs)
        # Turn off the line above, and on the line below if you wish to keep these extra kwargs.
        # self._process_kwargs(kwargs)

    def initialize(self):
        \"\"\"Initialize the link.

        :returns: status code of initialization
        :rtype: StatusCode
        \"\"\"
        return StatusCode.Success

    def execute(self):
        \"\"\"Execute the link.

        :returns: status code of execution
        :rtype: StatusCode
        \"\"\"
        settings = process_manager.service(ConfigObject)
        ds = process_manager.service(DataStore)

        # --- your algorithm code goes here
        self.logger.debug('Now executing link: {{link}}.', link=self.name)

        return StatusCode.Success

    def finalize(self):
        \"\"\"Finalize the link.

        :returns: status code of finalization
        :rtype: StatusCode
        \"\"\"
        # --- any code to finalize the link follows here

        return StatusCode.Success
"""
    file_name = link_name.lower()

    import_line = 'from .{file_name} import {link_name}'.format(file_name=file_name, link_name=link_name)
    create_file(path=link_dir,
                file_name='{file_name!s}.py'.format(file_name=file_name),
                content=template.format(link_name=link_name, date_generated=datetime.date.today()))
    if is_create_init:
        init_content = '# Created by Eskapade on {date!s}\n{import_line}\n'.format(date=datetime.date.today(),
                                                                                   import_line=import_line)
        create_file(path=link_dir,
                    file_name='__init__.py',
                    content=init_content)
    else:
        logger.info('Edit {link_dir}/__init__.py: add \"{link_name}\" to __all__ and add the line "{import_line}".'
                    .format(link_dir=link_dir, file_name=file_name, link_name=link_name, import_line=import_line))


def generate_macro(macro_dir,
                   macro_name,
                   link_module='eskapade.core_ops',
                   link_name='HelloWorld',
                   is_create_init=False):
    """Generate Eskapade macro.

    :param macro_dir: absolute path to a directory where the macro will be generated
    :param macro_name: name of the macro to generate
    :param link_module: module of a link to import
    :param link_name: name of the link to import
    :param is_create_init: whether to create __init__.py file or no
    """
    # Do not modify the indentation of template!
    template = """\"\"\"Project: Eskapade - A python-based package for data analysis.

Macro: {macro_name!s}

Created: {date_generated!s}

Description:
    Macro does ...(fill in short description here)

Authors:
    Your name(s) here

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
\"\"\"

from eskapade import process_manager, Chain, ConfigObject
from eskapade.logger import Logger, LogLevel

from links import {link_name!s}

logger = Logger()

logger.debug('Now parsing configuration file {macro_name!s}.')

# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = '{macro_name!s}'
settings['version'] = 0

# --- now set up the chains and links

ch = Chain('Start')
link = {link_name!s}()
link.logger.log_level = LogLevel.DEBUG
ch.add(link)

logger.debug('Done parsing configuration file {macro_name!s}.')
"""

    content = template.format(macro_name=macro_name,
                              date_generated=datetime.date.today(),
                              link_module=link_module,
                              link_name=link_name)
    if is_create_init:
        init_content = '# Created by Eskapade on {date!s}\n'.format(date=datetime.date.today())

        create_file(path=macro_dir,
                    file_name='__init__.py',
                    content=init_content)
    create_file(path=macro_dir,
                file_name='{macro_name!s}.py'.format(macro_name=macro_name),
                content=content)


def generate_notebook(notebook_dir, notebook_name, macro_path=None):
    """Generate Eskapade notebook.

    :param notebook_dir: absolute path to a directory where the notebook will be generated
    :param notebook_name: name of the notebook to generate
    :param macro_path: absolute path to a macro the notebook executes
    """
    import platform

    from eskapade import resources

    if macro_path:
        macro_path = "'{path}'".format(path=macro_path)
    else:
        macro_path = "resources.tutorial('tutorial_1.py')"

    with open(resources.template('notebook_template.ipynb')) as file:
        template = file.read()
        content = template.format(macro_path=macro_path,
                                  notebook_name=notebook_name,
                                  python_version=platform.python_version())
        create_file(path=notebook_dir,
                    file_name='{notebook_name!s}.ipynb'.format(notebook_name=notebook_name),
                    content=content)


def generate_configs(root_dir: str) -> None:
    """Generate default configs.

    :param str root_dir: Absolute path to package root directory.
    """
    config_dir = root_dir + '/config/'
    create_dir(config_dir)


def generate_setup(root_dir, package_name):
    """Generate project setup.py.

    :param root_dir: absolute path to an analysis project root dir
    :param package_name: package name
    """
    # Do not modify the indentation of template!
    template = """from setuptools import setup, find_packages

NAME = '{package_name}'

MAJOR = 1
REVISION = 0
PATCH = 0
DEV = True

VERSION = '{{major}}.{{revision}}.{{patch}}'.format(major=MAJOR, revision=REVISION, patch=PATCH)
FULL_VERSION = VERSION
if DEV:
    FULL_VERSION += '.dev'


def setup_package() -> None:
    \"\"\"The main setup method.

    It is responsible for setting up and installing the package.
    \"\"\"

    setup(name=NAME,
          version=VERSION,
          python_requires='>=3.5',
          package_dir={{'': '.'}},
          packages=find_packages(),
          install_requires=['eskapade']
          )


if __name__ == '__main__':
    setup_package()
"""
    content = template.format(package_name=package_name)
    create_file(path=root_dir, file_name='setup.py', content=content)



"""Project: Eskapade - A python-based package for data analysis.

Macro : esk304_df_boxplot

Created: 2017/02/23

Description:
    Macro shows how to boxplot the content of a dataframe in a nice summary
    pdf file.

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import analysis, visualization
from eskapade import process_manager
from eskapade.core import persistence
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk304_df_boxplot.')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk304_df_boxplot'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

msg = r"""

The plots and latex files produced by link df_summary can be found in dir:
{path}
"""
logger.info(msg, path=persistence.io_path('results_data', 'report'))

COLUMNS = ['var_a', 'var_b', 'var_c']
SIZE = 10000
VAR_LABELS = dict(var_a='Variable A', var_b='Variable B', var_c='Variable C')
VAR_UNITS = dict(var_b='m/s')
GEN_CONF = dict(var_a=dict(choice=['alpha', 'beta', 'gamma'], dtype=str), var_b=dict(mean=3., std=1.),
                var_c=dict(choice=['delta', 'epsilon', 'zeta', 'eta'], dtype=str))

#########################################################################################
# --- now set up the chains and links based on configuration flags

# create chains
data = Chain('Data')

# add data-generator link to "Data" chain
generator = analysis.BasicGenerator(name='Generate_data',
                                    key='data',
                                    columns=COLUMNS,
                                    size=SIZE,
                                    gen_config=GEN_CONF)
data.add(generator)

# add data-frame summary link to "Boxplot" chain
# can provide labels and units for the variables in the dataset, and set the statistics to print in output file
plot = Chain('BoxPlot')
box_plot = visualization.DfBoxplot(name='Create_stats_overview',
                                   read_key=generator.key,
                                   statistics=['count', 'mean', 'min', 'max', 'std'],
                                   var_labels=VAR_LABELS,
                                   var_units=VAR_UNITS,
                                   column='var_b',
                                   cause_columns=['var_a', 'var_c'],
                                   results_path=persistence.io_path('results_data', 'report'))
plot.add(box_plot)

#########################################################################################

logger.debug('Done parsing configuration file esk304_df_boxplot.')



"""Project: Eskapade - A python-based package for data analysis.

Classes: TimePeriod, FreqTimePeriod

Created: 2017/03/14

Description:
    Time period and time period with frequency.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

modification, are permitted according to the terms listed in the file
Redistribution and use in source and binary forms, with or without
LICENSE.
"""

import pandas as pd

from eskapade.core.mixin import ArgumentsMixin
from eskapade.logger import Logger


class TimePeriod(ArgumentsMixin):
    """Time period."""

    logger = Logger()

    def __init__(self, **kwargs):
        """Initialize TimePeriod instance."""
        pass

    def period_index(self, dt):
        """Get number of periods until date/time "dt".

        :param dt: specified date/time
        """
        self.logger.fatal('period_index method not implemented for {cls}; please implement derived class.',
                          cls=self.__class__.__name__)
        raise NotImplementedError('period_index function is not implemented.')

    @classmethod
    def parse_time_period(cls, period):
        """Try to parse specified time period.

        :param period: specified period
        """
        # catch single value
        if not isinstance(period, dict):
            period = dict(value=period)

        # try to parse specified period
        try:
            return pd.Timedelta(**period).delta
        except Exception as ex:
            cls.logger.fatal('Unable to parse period: {period!s}.', period=period)
            raise ex

    @classmethod
    def parse_date_time(cls, dt):
        """Try to parse specified date/time.

        :param dt: specified date/time
        """
        try:
            return pd.Timestamp(dt).value
        except Exception as ex:
            cls.logger.fatal('Unable to parse date/time: {dt!s}', dt=dt)
            raise ex


class UniformTsTimePeriod(TimePeriod):
    """Time period with offset."""

    def __init__(self, **kwargs):
        """Initialize TimePeriod instance."""
        super(UniformTsTimePeriod, self).__init__(**kwargs)

        # get parameters from arguments
        self._process_kwargs(kwargs, period='1d', offset=0)

        # check for extraneous keyword arguments
        self.check_extra_kwargs(kwargs)

        # check value of period
        if self.period <= 0:
            self.logger.fatal('Invalid time period specified: {period:d} ns.', period=self.period)
            raise AssertionError('Time period must be greater than zero.')

    def period_index(self, dt):
        """Get number of periods until date/time "dt" since "offset", given specified "period".

        :param dt: specified date/time
        """
        return (pd.Timestamp(dt).value - self.offset) // self.period

    @property
    def period(self):
        """Get period parameter."""
        return self._period

    @period.setter
    def period(self, period):
        """Set period parameter.

        :param period: specified period parameter
        """
        self._period = self.parse_time_period(period)

    @property
    def offset(self):
        """Get offset parameter."""
        return self._offset

    @offset.setter
    def offset(self, offset):
        """Set offset parameter.

        :param offset: specified offset parameter
        """
        self._offset = self.parse_date_time(offset)


class FreqTimePeriod(TimePeriod):
    """Time period with frequency."""

    def __init__(self, **kwargs):
        """Initialize TimePeriod instance."""
        super(FreqTimePeriod, self).__init__(**kwargs)

        # get parameters from arguments
        self._process_kwargs(kwargs, freq='D')

        # check for extraneous keyword arguments
        self.check_extra_kwargs(kwargs)

    def period_index(self, dt):
        """Return number of periods until date/time "dt" since 1970-01-01.

        :param dt: specified date/time parameter
        """
        return pd.Period(freq=self.freq, value=pd.Timestamp(dt)).ordinal

    def dt_string(self, period_index):
        """Convert period index into date/time string (start of period).

        :param int period_index: specified period index value.
        """
        return str(pd.Period(freq=self.freq, ordinal=period_index).start_time)

    @property
    def freq(self):
        """Return frequency."""
        return self._freq

    @freq.setter
    def freq(self, freq):
        """Try to construct a period object with specified "frequency".

        :param freq: specified frequency
        """
        try:
            per = pd.Period(freq=freq, value='1970-01-01')
        except Exception as ex:
            self.logger.fatal('Invalid frequency specified: {freq!s}.', freq=freq)
            raise ex

        # set "frequency"
        self._freq = per.freq



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk102_multiple_chains

Created: 2017/02/20

Description:
    Macro to illustrate the use of multiple chains

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import core_ops
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk102_multiple_chains')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk102_multiple_chains'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

settings['do_chain0'] = True
settings['do_chain1'] = True
settings['do_chain2'] = True

#########################################################################################
# --- now set up the chains and links based on configuration flags

# Three simple chains are set up.

if settings['do_chain0']:
    ch = Chain('Chain0')
    link = core_ops.HelloWorld(name='hello0')
    link.hello = 'Town'
    ch.add(link)

# adding more chains is as easy as creating multiple chain instances with unique names.

if settings['do_chain1']:
    ch = Chain('Chain1')
    link = core_ops.HelloWorld(name='hello1')
    link.hello = 'World'
    ch.add(link)

if settings['do_chain2']:
    ch = Chain('Chain2')
    link = core_ops.HelloWorld(name='hello2')
    link.hello = 'Universe'
    ch.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk102_multiple_chains')



"""Project: Eskapade - A python-based package for data analysis.

Created: 2016/11/08

Classes: ArgumentsMixin, TimerMixin

Description:
    Mixin classes:
        - ArgumentsMixin: processes/checks arguments and sets them as attributes
        - TimerMixin:     keeps track of execution time
        - ConfigMixin:    reads and handles settings from configuration files

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import configparser
import timeit


class ArgumentsMixin:
    """Mixin base class for argument parsing.

    Class allows attributes to be accessed as dict items.  Plus several
    argument processing helper functions.
    """

    def __contains__(self, key, *args, **kwargs):
        """Check if key is present.

        Dict style __contains__ function.
        Note: works for attributes, but does not work for properties.
        """
        if key in vars(self):
            return hasattr(self, key)
        super_meth = getattr(super(ArgumentsMixin, self), '__contains__', None)
        if not super_meth:
            return False
        return super_meth(key, *args, **kwargs)

    def __getitem__(self, key, *args, **kwargs):
        """Get item.

        Dict style __getitem__ function.
        Note: works for attributes, but does not work for properties.
        """
        if key in vars(self):
            return getattr(self, key)
        super_meth = getattr(super(ArgumentsMixin, self), '__getitem__', None)
        if not super_meth:
            raise KeyError('{!s} does not contain item "{!s}"'.format(self, key))
        return super_meth(key, *args, **kwargs)

    def __setitem__(self, key, item, *args, **kwargs):
        """Set item.

        Dict style __setitem__ function.
        """
        if isinstance(key, str):
            return setattr(self, key, item)
        super_meth = getattr(super(ArgumentsMixin, self), '__setitem__', None)
        if not super_meth:
            raise TypeError('item name must be a string, not "{}"'.format(type(key).__name__))
        return super_meth(key, item, *args, **kwargs)

    def _process_kwargs(self, kwargs, **name_val):
        """Add names from kwargs as attributes."""
        if not name_val:
            name_val = kwargs.copy()
        for name, def_val in name_val.items():
            setattr(self, name, kwargs.pop(name, def_val))

    def check_extra_kwargs(self, kwargs):
        """Check for residual kwargs."""
        for key in list(kwargs.keys()):
            if hasattr(self, key):
                kwargs.pop(key)
        if kwargs:
            raise RuntimeError('extraneous keyword arguments for {!s}: {!s}'.format(self, kwargs))

    def check_required_args(self, *arg_names):
        """Check if set of arguments exists as attributes."""
        for arg_name in arg_names:
            if not hasattr(self, arg_name):
                raise AttributeError('argument "{}" not set for {!s}'.format(arg_name, self))

    def check_arg_vals(self, *arg_names, allow_none=False):
        """Check if set of arguments exists as attributes and values."""
        self.check_required_args(*arg_names)
        for arg_name in arg_names:
            attr = getattr(self, arg_name)
            if not (allow_none and attr is None) and not bool(attr):
                raise ValueError('argument "{}" of {!s} has no value'.format(arg_name, self))

    def check_arg_opts(self, allow_none=False, **name_vals):
        """Check if argument values are in set of options."""
        self.check_required_args(*tuple(name_vals.keys()))
        for arg_name, arg_opts in name_vals.items():
            attr = getattr(self, arg_name)
            if not (allow_none and attr is None) and attr not in arg_opts:
                raise ValueError('invalid value for argument "{}" of {!s}: "{!s}" (options are {!s})'
                                 .format(arg_name, self, attr, tuple(arg_opts)))

    def check_arg_types(self, recurse=False, allow_none=False, **name_type):
        """Check if set of arguments has correct types."""
        def check_attr(self, attr, a_name, a_type):
            """Check if argument has correct type."""
            if allow_none and attr is None:
                return
            if not isinstance(attr, a_type):
                if recurse and hasattr(attr, '__iter__'):
                    for a in attr:
                        check_attr(self, a, a_name, a_type)
                else:
                    raise TypeError('type of (element of) argument "{}" of {!s} is "{}" ("{}" required)'
                                    .format(a_name, self, type(attr).__name__, a_type.__name__))

        self.check_required_args(*tuple(name_type.keys()))
        for arg_name, arg_type in name_type.items():
            check_attr(self, getattr(self, arg_name), arg_name, arg_type)

    def check_arg_iters(self, *arg_names, allow_none=False):
        """Check if set of arguments has iterators."""
        self.check_required_args(*arg_names)
        for arg_name in arg_names:
            attr = getattr(self, arg_name)
            if allow_none and attr is None:
                continue
            if not hasattr(getattr(self, arg_name), '__iter__'):
                raise TypeError('argument "{}" of {!s} is not iterable'.format(arg_name, self))

    def check_arg_callable(self, *arg_names, allow_none=False):
        """Check if set of arguments has iterators."""
        self.check_required_args(*arg_names)
        for arg_name in arg_names:
            attr = getattr(self, arg_name)
            if allow_none and attr is None:
                continue
            if not callable(getattr(self, arg_name)):
                raise TypeError('argument "{}" of {!s} is not callable'.format(arg_name, self))


class TimerMixin:
    """Mixin base class for timing."""

    def __init__(self):
        """Initialize timer."""
        self._start_time = 0.
        self._stop_time = 0.
        self._total_time = 0.

    def start_timer(self):
        """Start run timer.

        Start the timer. The timer is used to compute the run time.  The
        returned timer start value has an undefined reference and should,
        therefore, only be compared to other timer values.

        :returns: start time in seconds
        :rtype: float
        """
        self._start_time = timeit.default_timer()
        return self._start_time

    def stop_timer(self, start_time=None):
        """Stop the run timer.

        Stop the timer.  The timer is used to compute the run time.  The
        elapsed time since the timer start is returned.

        :param float start_time: function start_time input
        :returns: time difference with start in seconds
        :rtype: float
        """
        self._stop_time = timeit.default_timer()

        diff_time = self._stop_time - (start_time if start_time is not None else self._start_time)
        self._total_time += diff_time

        return diff_time

    def total_time(self):
        """Return the total run time.

        :returns: total time in seconds
        :rtype: float
        """
        return self._total_time


class ConfigMixin:
    """Mixin base class for configuration settings."""

    def __init__(self, config_path=None):
        """Initialize config settings.

        :param str config_path: path of configuration file
        """
        self._config_path = str(config_path) if config_path else ''
        self._config = None

    @property
    def config_path(self):
        """Path of configuration file."""
        return self._config_path

    @config_path.setter
    def config_path(self, path):
        """Set path of configuration file.

        :param str path: path to be set
        :raises ValueError: if path is not a string or has no value
        """
        path = str(path) if path else ''
        if not path:
            raise ValueError('no value specified for config-file path')
        self._config_path = path

    def get_config(self, config_path=None):
        """Get settings from configuration file.

        Read and return the configuration settings from a configuration file.
        If the path of this file is not specified as an argument, the value of
        the "config_path" property is used.  If the file has already been read,
        return previous settings.

        :param str config_path: path of configuration file
        :returns: configuration settings read from file
        :rtype: configparser.ConfigParser
        :raises RuntimeError: if config_path is not set
        """
        if not self._config:
            # set config-path attribute
            if config_path:
                self.config_path = config_path
            elif not self.config_path:
                raise RuntimeError('no configuration-file path set')

            # read configuration file
            self._config = configparser.ConfigParser()
            self._config.read(self.config_path)

        return self._config

    def reset_config(self):
        """Remove previously read settings."""
        self._config = None



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk105_datastore_pickling

Created: 2017/02/20

Description:
    Macro illustrates how to start running at any particular chain by
    picking up the datastore and configuration objects from the
    previous chain.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject
from eskapade import process_manager, resources
from eskapade.logger import Logger

logger = Logger()

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# turning on this flag, the process manager starts off at chain3,
# and does so by reading in the datastore written out after chain 2.

settings = process_manager.service(ConfigObject)
settings['beginWithChain'] = 'chain3'

msg = r"""

--> Make sure to run this macro after running macro
    esk105_B_store_each_chain.py

By default, the process manager starts the execution of the first
configured chain and end with the last one.  If the run-process services
are written out after the execution of each chain (--store-all), it is
possible to execute one particular chain.  This is done by picking up
the services written out by the previous chain.  This is a nice feature
for debugging and developing purposes.

From the command line use the options:

-b BEGIN_WITH_CHAIN
-e END_WITH_CHAIN
-s SINGLE_CHAIN
"""
logger.info(msg)

#########################################################################################
# --- now parse the follow-up macro

# the flag doNotStoreResults is picked up when parsing the following macro
process_manager.execute_macro(resources.tutorial('esk105_datastore_pickling.py'))



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk107_chain_looper

Created: 2017/02/17

Description:
    Macro to that illustrates how to repeat the execution of a chain.
    Follow-up example in macro: esk209_readdata_itr.py

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import core_ops
from eskapade import process_manager
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk107_chain_looper')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk107_chain_looper'
settings['version'] = 0

#########################################################################################
# --- Analysis configuration flags.
#     E.g. use these flags turn on or off certain chains with links.
#     by default all set to false, unless already configured in
#     configobject or vars()

# turn on/off the example
settings['do_example'] = True

#########################################################################################
# --- now set up the chains and links, based on configuration flags


# --- example loops over the first chain 10 times.

if settings['do_example']:
    # --- a loop is set up in the chain MyChain.
    #     we iterate over the chain until the link RepeatChain is done.
    #     then move on to the next chain (Overview)
    ch = Chain('MyChain')

    link = core_ops.HelloWorld(name='HelloWorld')
    link.logger.log_level = LogLevel.DEBUG
    ch.add(link)

    # --- this link sends out a signal to repeat the execution of the chain.
    #     It serves as the 'continue' statement of the loop.
    #     go back to start of the chain until counter reaches 10.
    repeater = core_ops.RepeatChain()
    # repeat max of 10 times
    repeater.maxcount = 10
    repeater.logger.log_level = LogLevel.DEBUG
    ch.add(repeater)

# --- print contents of the datastore.
#    which in this case is empty.
overview = Chain('Overview')
pds = core_ops.PrintDs(name='End')
overview.add(pds)

#########################################################################################

logger.debug('Done parsing configuration file esk107_chain_looper')



"""Project: Eskapade - A python-based package for data analysis.

Class: RecordFactorizer

Created: 2016/11/08

Description:
    Algorithm to perform the factorization of an input column
    of an input dataframe.
    E.g. a columnn x with values 'apple', 'tree', 'pear', 'apple', 'pear'
    is tranformed into columns x with values 0, 1, 2, 0, 2, etc.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy
import fnmatch

import pandas as pd
from pandas import DataFrame

try:
    from pandas.types.dtypes import CategoricalDtypeType
except ImportError:
    from pandas.core.dtypes.dtypes import CategoricalDtypeType

from eskapade import process_manager
from eskapade import Link
from eskapade import StatusCode
from eskapade import DataStore


class RecordFactorizer(Link):
    """Factorize data-frame columns.

    Perform factorization of input column of an input dataframe.  E.g. a
    columnn x with values 'apple', 'tree', 'pear', 'apple', 'pear' is
    tranformed into columns x with values 0, 1, 2, 0, 2, etc.  Resulting
    dataset stored as new dataset.  Alternatively, map transformed columns
    back to orginal format.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        Store and do basic check on the attributes of link RecordFactorizer

        :param str read_key: key to read dataframe from the data store. Dataframe of records that is to be transformed.
        :param list columns: list of columns that are to be factorized
        :param bool inplace: replace original columns. Default is False. Overwrites store_key to read_key.
        :param bool convert_all_categories: if true, convert all catergory observables. Default is false.
        :param bool convert_all_booleans: if true, convert all boolean observables. Default is false.
        :param dict map_to_original: dictiorary or key to dictionary to map back factorized columns to original.
                                     map_to_original is a dict of dicts, one dict for each column.
        :param str store_key: store key of output dataFrame. Default is read_key + '_fact'. (optional)
        :param str sk_map_to_original: store key of dictiorary to map factorized columns to original.
                                       Default is 'key' + '_' + store_key + '_to_original'. (optional)
        :param str sk_map_to_factorized: store key of dictiorary to map original to factorized columns.
                                         Default is 'key' + '_' + read_key + '_to_factorized'. (optional)
        """
        Link.__init__(self, kwargs.pop('name', 'RecordFactorizer'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             read_key='',
                             columns=[],
                             convert_all_categories=False,
                             convert_all_booleans=False,
                             store_key='',
                             sk_map_to_original='',
                             sk_map_to_factorized='',
                             map_to_original={},
                             inplace=False)

        # check residual kwargs. exit if any present
        self.check_extra_kwargs(kwargs)

        # map to factorized filled during execution
        self._mtf = {}
        # map to original, dict set below
        self._mto = {}

    def initialize(self):
        """Initialize the link.

        Initialize and (further) check the assigned attributes of
        the RecordFactorizer
        """
        self.check_arg_types(read_key=str, store_key=str, sk_map_to_original=str, sk_map_to_factorized=str)
        self.check_arg_types(recurse=True, allow_none=True, columns=str)
        self.check_arg_vals('read_key')

        if self.inplace:
            self.store_key = self.read_key
            self.logger.debug('store_key has been set to read_key "{key}."', key=self.store_key)

        if not self.store_key:
            self.store_key = self.read_key + '_fact'
            self.logger.debug('store_key has been set to "{key}".', key=self.store_key)
        if not self.sk_map_to_original:
            self.sk_map_to_original = 'map_' + self.store_key + '_to_original'
            self.logger.debug('Storage key <sk_map_to_original> has been set to "{key}".', key=self.sk_map_to_original)
        if not self.sk_map_to_factorized:
            self.sk_map_to_factorized = 'map_' + self.read_key + '_to_factorized'
            self.logger.debug('Storage key <sk_map_to_factorized> has been set to "{key}".',
                              key=self.sk_map_to_factorized)

        if self.map_to_original and not isinstance(self.map_to_original, str) \
                and not isinstance(self.map_to_original, dict):
            raise TypeError('map_to_original needs to be a dict or string (to fetch a dict from the datastore)')

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Perform factorization input columns 'columns' of input dataframe.
        Resulting dataset stored as new dataset.
        Alternatively, map transformed columns back to orginal format.
        """
        ds = process_manager.service(DataStore)

        # basic checks on contensts of the data frame
        if self.read_key not in ds:
            raise KeyError('Key "{}" not in DataStore'.format(self.read_key))
        df = ds[self.read_key]
        if not isinstance(df, DataFrame):
            raise TypeError('retrieved object not of type pandas DataFrame')
        if len(df.index) == 0:
            raise AssertionError('dataframe "{}" is empty'.format(self.read_key))
        # match all columns/pattern in self.columns to df.columns
        matched_columns = []
        for c in self.columns:
            match_c = fnmatch.filter(df.columns, c)
            if not match_c:
                raise AssertionError('column or pattern "{}" not present in input data frame'.format(c))
            matched_columns += match_c
        self.columns = matched_columns
        # convert booleans and categorical observables?
        for c in set(df.columns).difference(self.columns):
            dt = df[c].dtype
            if (issubclass(dt.type, CategoricalDtypeType) and self.convert_all_categories) \
                    or (dt == 'bool' and self.convert_all_booleans):
                self.columns.append(c)
        # retrieve map_to_original from ds
        if self.map_to_original:
            if isinstance(self.map_to_original, str):
                assert len(self.map_to_original), 'map_to_original needs to be a filled string.'
                assert self.map_to_original in ds, 'map_to_original key not found in datastore.'
                self._mto = ds[self.map_to_original]
            elif isinstance(self.map_to_original, dict):
                self._mto = self.map_to_original
            assert isinstance(self._mto, dict), 'map_to_original needs to be a dict.'

        # 1. do factorization for all specified columns
        if not self.map_to_original:
            df_fact = df if self.inplace else pd.DataFrame(index=df.index)
            for c in self.columns:
                self.logger.debug('Factorizing column "{column}" of dataframe "{key}".',
                                  column=c, key=self.read_key)
                labels, unique = df[c].factorize()
                df_fact[c] = labels
                self._mto[c] = dict((i, v) for i, v in enumerate(unique))
                self._mtf[c] = dict((v, i) for i, v in enumerate(unique))
            # store the mapping here
            ds[self.sk_map_to_original] = self._mto
            ds[self.sk_map_to_factorized] = self._mtf
        # 2. do the mapping back to original format
        else:
            # pandas replace() will not do transformations that are identical, including int 0/1 to bool.
            # skip those column tranformations
            mto = copy.copy(self._mto)
            for c, c_mto in self._mto.items():
                k = list(c_mto.keys())
                v = list(c_mto.values())
                if set(k) & set(v):
                    # true in case of indentical transformation
                    self.logger.debug('Identical transformation for column "{column}". Skipping column.', column=c)
                    del mto[c]
            # refactorizing
            self.logger.debug('Refactorizing columns {columns} of dataframe "{key}".',
                              columns=list(mto.keys()), key=self.read_key)
            df_fact = df.replace(mto, inplace=self.inplace)
            if self.inplace:
                df_fact = df

        # storage
        ds[self.store_key] = df_fact

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Class: BasicGenerator

Created: 2017/02/26

Description:
    Link to generate random data with basic distributions

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import numpy as np
import pandas as pd

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class BasicGenerator(Link):
    """Generate data with basic distributions."""

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str key: key of output data in data store
        :param list columns: output column names
        :param int size: number of variable values
        :param dict gen_config: generator configuration for each variable
        :param int gen_seed: generator random seed
        """
        # initialize Link
        Link.__init__(self, kwargs.pop('name', 'basic_generator'))

        # process keyword arguments
        self._process_kwargs(kwargs, key='', columns=None, size=1, gen_config=None, gen_seed=1)
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link."""
        # check input arguments
        self.check_arg_vals('key', 'size', 'columns')
        self.check_arg_types(key=str, size=int, gen_seed=int)
        self.check_arg_types(recurse=True, allow_none=True, columns=str, gen_config=str)

        # set generator seed
        np.random.seed(self.gen_seed)

        return StatusCode.Success

    def execute(self):
        """Execute the link."""
        # generate data
        self.logger.debug('Generating {n:d} rows for columns [{columns}].',
                          n=self.size, columns=', '.join('"{}"'.format(c) for c in self.columns))
        data = {}
        for col in self.columns:
            # get generator configuration for this variable
            conf = self.gen_config.get(col, {}) if self.gen_config else {}
            mu = conf.get('mean', 0.)
            sigma = conf.get('std', 1.)
            dtype = conf.get('dtype', float)
            choice = conf.get('choice', ['a', 'b', 'c'])
            choice_prob = conf.get('choice_prob', None)

            # generate
            if dtype == str:
                data[col] = np.random.choice(choice, size=self.size, p=choice_prob)
            else:
                data[col] = np.random.normal(loc=mu, scale=sigma, size=self.size).astype(dtype)

        # create data frame
        process_manager.service(DataStore)[self.key] = pd.DataFrame(data=data, columns=self.columns)

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Module: helpers

Created: 2017/05/17

Description:
    Helper functions for Eskapade

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade.logger import Logger

logger = Logger()


def obj_repr(obj):
    """Get generic string representation of object."""
    return object.__repr__(obj)


def apply_transform_funcs(obj, trans_funcs, func_args=None, func_kwargs=None):
    """Transform object by applying transformation functions.

    Transformation functions are applied sequentially to the output of the
    previous function, starting with the specified object.  The final
    resulting object is returned.

    Functions are specified with the "trans_funcs" argument.  This is an
    iterable of either the functions themselves or (function, positional
    arguments, keyword arguments) combinations.  The function arguments can
    also be specified by the "func_args" and "func_kwargs" arguments.  The
    functions are called with the object to be transformed as a first
    argument, followed by the specified positional and keyword arguments.

    A specified function can be either a callable object or a string.  The
    latter is interpreted as a method of the type of the object on which the
    function is applied.

    :param obj: object to transform
    :param iterable trans_funcs: functions to apply, specified the function or a (function, args, kwargs) tuple
    :param dict func_args: function positional arguments, specified as (function, arguments tuple) pairs
    :param dict func_kwargs: function keyword arguments, specified as (function, kwargs dict) pairs
    :returns: transformed object
    """
    # process input functions
    trans_funcs = process_transform_funcs(trans_funcs, func_args, func_kwargs)

    # loop over specified functions
    for func, args, kwargs in trans_funcs:
        # get member function of object type if string is specified
        if isinstance(func, str):
            # get function
            try:
                func = getattr(type(obj), func)
            except AttributeError:
                raise AttributeError('no member function "{0:s}" for object "{1:s}"'.format(func, obj_repr(obj)))

            # check if function is callable
            if not callable(func):
                raise TypeError('function "{0!s}" for object "{1:s}" not callable'.format(func, obj_repr(obj)))

        # apply function
        logger.debug('Applying transformation function "{func!s}" to "{object}":', func=func, object=obj_repr(obj))
        logger.debug('    args:   {args!s}', args=args)
        logger.debug('    kwargs: {kwargs!s}', kwargs=kwargs)
        obj = func(obj, *args, **kwargs)

    return obj


def process_transform_funcs(trans_funcs, func_args=None, func_kwargs=None):
    """Process input of the apply_transform_funcs function.

    :param iterable trans_funcs: functions to apply, specified the function or a (function, args, kwargs) tuple
    :param dict func_args: function positional arguments, specified as (function, arguments tuple) pairs
    :param dict func_kwargs: function keyword arguments, specified as (function, kwargs dict) pairs
    :returns: transformation functions for apply_transform_funcs function
    :rtype: list
    """
    # copy function arguments and make sure they are in dictionaries
    func_args = dict(func_args) if func_args else {}
    func_kwargs = dict(func_kwargs) if func_kwargs else {}

    # loop over specified functions
    proc_funcs = []
    for func_spec in trans_funcs:
        # expand (func, args, kwargs) combinations
        try:
            func, args, kwargs = (func_spec, None, None) if isinstance(func_spec, str) else func_spec
        except TypeError:
            func, args, kwargs = func_spec, None, None
        except ValueError:
            raise ValueError('expected (func, args, kwargs) combination (got {!s})'.format(func_spec))

        # get positional arguments
        args_ = func_args.pop(func, None)
        if args is not None and args_ is not None:
            raise RuntimeError('arguments for "{!s}" in both "trans_funcs" and "func_args"'.format(func))
        args = tuple(args) if args is not None else tuple(args_) if args_ is not None else ()

        # get keyword arguments
        kwargs_ = func_kwargs.pop(func, None)
        if kwargs is not None and kwargs_ is not None:
            raise RuntimeError('keyword arguments for "{!s}" in both "trans_funcs" and "func_kwargs"'.format(func))
        kwargs = dict(kwargs) if kwargs is not None else dict(kwargs_) if kwargs_ is not None else {}

        # check if function is callable
        if not (isinstance(func, str) or callable(func)):
            raise TypeError('function "{!s}" is not callable'.format(func))

        # append function specification
        proc_funcs.append((func, args, kwargs))

    # check if all specified arguments were used
    if func_args:
        raise ValueError('unused function arguments specified: {!s}'.format(func_args))
    if func_kwargs:
        raise ValueError('unused function keyword arguments specified: {!s}'.format(func_kwargs))

    return proc_funcs



"""Project: Eskapade - A python-based package for data analysis.

Class: ValueCounter

Created: 2017/03/02

Description:
    Algorithm to do value_counts() on single columns of a pandas
    dataframe, or groupby().size() on multiple columns, both returned
    as dictionaries. It is possible to do cleaning of these dicts by
    rejecting certain keys or removing inconsistent data types.
    Numeric and timestamp columns are converted to bin indices before
    the binning is applied.
    Results are stored as 1D Histograms or as ValueCounts objects.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from collections import Counter

import numpy as np

from eskapade import process_manager, DataStore
from eskapade.analysis import histogram_filling as hf
from eskapade.analysis.histogram import Histogram, ValueCounts
from eskapade.analysis.histogram_filling import HistogramFillerBase


class ValueCounter(HistogramFillerBase):
    """Count values in Pandas data frame.

    ValueCounter does value_counts() on single columns of a pandas
    dataframe, or groupby().size() on multiple columns.  Results of both are
    returned as same-style dictionaries.

    Numeric and timestamp columns are converted to bin indices before the
    binning is applied.  The binning can be provided as input.

    It is possible to do cleaning of these dicts by rejecting certain keys
    or removing inconsistent data types.  Results are stored as 1D
    Histograms or as ValueCounts objects.

    Example is available in: tutorials/esk302_histogram_filling_plotting.py
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str read_key: key of input data to read from data store
        :param str store_key_counts: key of output data to store ValueCounts objects in data store
        :param str store_key_hists: key of output data to store histograms in data store
        :param list columns: columns to pick up from input data (default is all columns)
        :param dict bin_specs: dictionaries used for rebinning numeric or timestamp columns

        Example bin_specs dictionary is:

        >>> bin_specs = {'x': {'bin_width': 1, 'bin_offset': 0},
        >>>              'y': {'bin_edges': [0, 2, 3, 4, 5, 7, 8]},
        >>>              'date': {'bin_width': np.timedelta64(30, 'D'),
        >>>                       'bin_offset': np.datetime64('2010-01-04')}}

        :param dict var_dtype: dict of datatypes of the columns to study from dataframe. If not provided, try to
                               determine datatypes directy from dataframe.
        :param bool store_at_finalize: Store histograms and/or ValueCount object in datastore at finalize(), not at
               execute(). Useful when looping over datasets. Default is False.
        :param bool drop_inconsistent_key_types: cleanup histograms and/or ValueCount objects by removing alls
               bins/keys with inconsistent datatypes. By default compare with data types in var_dtype dictionary.
        :param drop_keys dict: dictionary used for dropping specific keys from created value_counts dictionaries

        Example drop_keys dictionary is:

        >>> drop_keys = {'x': [1, 4, 8, 19],
        >>>              'y': ['apple', 'pear', 'tomato'],
        >>>              'x:y': [(1, 'apple'), (19, 'tomato')]}
        """
        # initialize Link, pass name from kwargs
        if 'name' not in kwargs:
            kwargs['name'] = 'ValueCounter'
        HistogramFillerBase.__init__(self, **kwargs)

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from
        # kwargs.
        self._process_kwargs(kwargs,
                             store_key_counts=None,
                             store_key_hists=None,
                             drop_inconsistent_key_types=True)

        # these get filled during execution
        self._counts = {}
        self._valcnts = {}

    def initialize(self):
        """Initialize the link."""
        # check basic attribute settings
        if self.store_key_hists is not None:
            assert isinstance(self.store_key_hists, str) and len(self.store_key_hists), \
                'store_key_hists has not been set to string'
        if self.store_key_counts is not None:
            assert isinstance(self.store_key_counts, str) and len(self.store_key_counts), \
                'store_key_counts has not been set to string'
        assert self.store_key_hists is not None or self.store_key_counts is not None, \
            'no store key has been set'

        return HistogramFillerBase.initialize(self)

    def process_columns(self, df):
        """Process columns before histogram filling.

        Specifically, convert timestamp columns to integers
        and numeric variables are converted to indices

        :param df: input (pandas) data frame
        :returns: output (pandas) data frame with converted timestamp columns
        :rtype: pandas DataFrame
        """
        # timestamp variables are converted to ns here
        # make temp df for value counting (used below)

        idf = df[self.str_cols].copy(deep=False)
        for col in self.dt_cols:
            self.logger.debug('Converting column "{column}" of type "{type}" to nanosec.',
                              column=col, type=self.var_dtype[col])
            idf[col] = df[col].apply(hf.to_ns)

        # numerical variables are converted to indices here
        for col in self.num_cols + self.dt_cols:
            self.logger.debug('Converting column "{column}" of type "{type}" to index.',
                              column=col, type=self.var_dtype[col])
            # find column specific bin_specs. if not found, use dict of default
            # values.
            dt = df[col].dtype
            is_number = isinstance(dt.type(), np.number)
            is_timestamp = isinstance(dt.type(), np.datetime64)
            sf = idf if is_timestamp else df
            bin_specs = self.bin_specs.get(col, self._unit_bin_specs if is_number else self._unit_timestamp_specs)
            idf[col] = sf[col].apply(hf.value_to_bin_index, **bin_specs)

        return idf

    def fill_histogram(self, idf, columns):
        """Fill input histogram with column(s) of input dataframe.

        :param idf: input data frame used for filling histogram
        :param list columns: histogram column(s)
        """
        name = ':'.join(columns)
        if name not in self._counts:
            # create an (empty) value counts dict
            self._counts[name] = Counter()
        # value_counts() is faster than groupby().size(), but only works for series (1d).
        # else use groupby() for multi-dimensions
        g = idf.groupby(by=columns).size() if len(columns) > 1 else idf[columns[0]].value_counts()
        counts = Counter(g.to_dict())
        # remove specific keys from histogram before merging, if so requested
        counts = self.drop_requested_keys(name, counts)
        self._counts[name].update(counts)

    def process_and_store(self):
        """Make, clean, and store ValueCount objects."""
        # nothing to do?
        if self.store_key_hists is None and self.store_key_counts is None:
            return

        ds = process_manager.service(DataStore)

        # 1. construct value counts
        for col in self.columns:
            name = ':'.join(col)
            vc = ValueCounts(col, col, self._counts[name])
            # remove all items from Counters where the key is not of correct datatype.
            # e.g. in Counter dict of ints, remove any non-ints that may arise
            # from dq issues.
            if self.drop_inconsistent_key_types:
                vc = self.drop_inconsistent_keys(col, vc)
            self._valcnts[name] = vc

        if self.store_key_counts is not None:
            ds[self.store_key_counts] = self._valcnts

        # 2. construct hists from value counts
        if self.store_key_hists is None:
            return

        for col in self.columns:
            if len(col) != 1:
                continue
            name = ':'.join(col)
            dt = np.dtype(self.var_dtype[name]).type()
            is_number = isinstance(dt, np.number)
            is_timestamp = isinstance(dt, np.datetime64)

            # bin_specs is used for converting index back to original var in
            # histogram class.
            bin_specs = {}
            if is_number:
                bin_specs = self.bin_specs.get(name, self._unit_bin_specs)
            elif is_timestamp:
                bin_specs = self.bin_specs.get(name, self._unit_timestamp_specs)
            h = Histogram(self._valcnts[name], variable=name, datatype=self.var_dtype[name],
                          bin_specs=bin_specs)
            self._hists[name] = h
        # and store
        ds[self.store_key_hists] = self._hists

    def drop_inconsistent_keys(self, columns, obj):
        """Drop inconsistent keys.

        Drop inconsistent keys from a ValueCounts or Histogram object.

        :param list columns: columns key to retrieve desired datatypes
        :param object obj: ValueCounts or Histogram object to drop inconsistent keys from
        """
        # has array been converted first? if so, set correct comparison
        # datatype
        comp_dtype = []
        for col in columns:
            dt = np.dtype(self.var_dtype[col]).type()
            is_converted = isinstance(dt, (np.number, np.datetime64))
            if is_converted:
                comp_dtype.append(np.int64)
            else:
                comp_dtype.append(self.var_dtype[col])
        # keep only keys of types in comp_dtype
        obj.remove_keys_of_inconsistent_type(prefered_key_type=comp_dtype)
        return obj

    def finalize(self):
        """Finalize ValueCounter."""
        status = HistogramFillerBase.finalize(self)
        # cleanup
        if self.store_key_counts is None:
            del self._valcnts
        if self.store_key_hists is None:
            del self._hists

        return status



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk110_code_profiling

Created: 2017/02/26

Description:
    Macro to demo how to run eskapade with code profiling turned on

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk110_code_profiling.')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk110_code_profiling'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

msg = r"""

Your can profile the speed of your analysis functions by running with the cmd line option: -P
You will need to select the order in which the profile output is shown on screen.
For example: -P cumulative

After running, this prints out a long list of all functions called, including the time it
took to run each of of them, sorted in the way you have specified.
"""
logger.info(msg)

# turn on code profiling in the ConfigObject
# turn off this line if you want to experiment with the profiling from the cmd line!
settings['doCodeProfiling'] = 'cumulative'

#########################################################################################
# --- now set up the chains and links based on configuration flags

# Just look at the code profiling output on the screen!
# The output is sorted by cumulative processing time.

# For more documentation on code profiling, see:
# https://docs.python.org/2/library/profile.html

#########################################################################################

logger.debug('Done parsing configuration file esk110_code_profiling.')



"""Project: Eskapade - A python-based package for data analysis.

Class: RecordVectorizer

Created: 2016/11/08

Description:
    Algorithm to perform the vectorization of an input column
    of an input dataframe.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from functools import reduce

import pandas as pd
from pandas import DataFrame

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class RecordVectorizer(Link):
    """Vectorize data-frame columns.

    Perform vectorization of input column of an input dataframe.  E.g. a
    columnn x with values 1, 2 is tranformed into columns x_1 and x_2, with
    values True or False assigned per record.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        Store and do basic check on the attributes of link RecordVectorizer.

        :param str read_key: key to read dataframe from the data store. Dataframe of records that is to be transformed.
        :param list columns: list of columns that are to be vectorized
        :param str store_key: store key of output dataFrame. Default is read_key + '_vectorized'. (optional)
        :param dict column_compare_with: dict of unique items per column with which column values are compared.
               If not given, this is derived automatically from the column. (optional)
        :param type astype: store answer of comparison of column with value as certain type. Default is bool. (optional)
        """
        Link.__init__(self, kwargs.pop('name', 'RecordVectorizer'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             read_key='',
                             store_key=None,
                             columns=[],
                             column_compare_with={},
                             astype=bool)

        # check residual kwargs. exit if any present
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link.

        Initialize and (further) check the assigned attributes of
        RecordVectorizer.
        """
        self.check_arg_types(read_key=str)
        self.check_arg_types(recurse=True, allow_none=True, columns=str)
        self.check_arg_vals('read_key')

        if self.store_key is None:
            self.store_key = self.read_key + '_vectorized'
            self.logger.info('Store key was empty, has been set to "{key}".', key=self.store_key)

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Perform vectorization input column 'column' of input dataframe.
        Resulting dataset stored as new dataset.
        """
        ds = process_manager.service(DataStore)

        # basic checks on contensts of the data frame
        if self.read_key not in ds:
            raise KeyError('key "{}" not in data store'.format(self.read_key))
        df = ds[self.read_key]
        if not isinstance(df, DataFrame):
            raise TypeError('retrieved object not of type pandas DataFrame')
        if len(df.index) == 0:
            raise AssertionError('dataframe "{}" is empty'.format(self.read_key))
        for c in self.columns:
            if c not in df.columns:
                raise AssertionError('column "{}" not present in input data frame'.format(c))

        # checks of column_compare_with
        if isinstance(self.column_compare_with, str) and len(self.column_compare_with):
            if self.column_compare_with not in ds:
                raise KeyError('column compare with "{}" not found in data store'.format(self.column_compare_with))
            self.column_compare_with = df[self.column_compare_with]
        if not isinstance(self.column_compare_with, dict):
            raise RuntimeError('column compare dict not set correctly')
        for c in self.columns:
            if c not in self.column_compare_with:
                self.column_compare_with[c] = df[c].unique()
            elif not isinstance(self.column_compare_with[c], list):
                raise TypeError('column "{}" needs to be compared with list of values'.format(c))

        # do vectorization for all columns, then merge
        dfs = [record_vectorizer(df, c, self.column_compare_with[c], self.astype) for c in self.columns]
        df_final = reduce(lambda left, right: pd.merge(left, right, left_index=True, right_index=True), dfs)

        ds[self.store_key] = df_final

        return StatusCode.Success


def record_vectorizer(df, column_to_vectorize, column_compare_set, astype=bool):
    """Vectorize data-frame column.

    Takes the new record that is already transformed and vectorizes the
    given columns.

    :param df: dataframe of the new record to vectorize
    :param str column_to_vectorize: string, column in the new record to vectorize.
    :param list column_compare_set: list of values to compare the column with.
    :returns: dataframe of the new records.
    """
    df_new = pd.DataFrame(index=df.index)
    for val in column_compare_set:
        newcol = column_to_vectorize + '_' + str(val)
        df_new[newcol] = (df[column_to_vectorize] == val).astype(astype)

    return df_new



"""Project: Eskapade - A python-based package for data analysis.

Class: PrintDs

Created: 2016/11/08

Description:
    Algorithm to print the content of the datastore.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import process_manager, Link, DataStore, StatusCode


class PrintDs(Link):
    """Print the content of the datastore."""

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param list keys: keys of items to print explicitly.
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'PrintDs'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs, keys=[])

        # check residual kwargs. exit if any present.
        self.check_extra_kwargs(kwargs)

    def execute(self):
        """Execute the link.

        Print overview of the datastore in current state.
        """
        ds = process_manager.service(DataStore)
        ds.Print()

        if not self.keys:
            return StatusCode.Success
        self.logger.info("*-------------------------------------------------*")
        for k in sorted(self.keys):
            if k not in ds:
                self.logger.warning('Datastore does not contain key "{key}". Skipping.', key=k)
                continue
            item = ds[k]
            line = "  {key} = {value}".format(key=k, value=type(item) if not hasattr(item, '__str__') else str(item))
            self.logger.info(line)
        self.logger.info("*-------------------------------------------------*")
        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Created: 2016/11/08

Description:
    Functions for running and resetting Eskapade machinery

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import eskapade.utils
from eskapade.core.process_manager import process_manager
from eskapade.core.process_services import ConfigObject
from eskapade.logger import Logger

logger = Logger()


def reset_eskapade(skip_config=False):
    """Reset Eskapade objects.

    :param bool skip_config: skip reset of configuration object
    """
    settings = process_manager.service(ConfigObject)
    process_manager.reset()
    if skip_config:
        process_manager.service(settings)


def eskapade_run(settings=None):
    """Run Eskapade.

    This function is called in the script eskapade_run when run
    from the cmd line.  The working principle of Eskapade is to run chains of
    custom code chunks (so-called links).

    Each chain should have a specific purpose, for example pre-processing
    incoming data, booking and/or training predictive algorithms, validating
    these predictive algorithms, evaluating the algorithms.

    By using this principle, links can be easily reused in future projects.

    :param ConfigObject settings: analysis settings
    :return: status of the execution
    :rtype: StatusCode
    """
    # get config object from process manager
    if settings:
        # use supplied settings as config service in process manager
        process_manager.remove_service(ConfigObject, silent=True)
        process_manager.service(settings)
    settings = process_manager.service(ConfigObject)

    def message(msg):
        width = 80
        fence = '*'
        logger.info(fence * width)
        logger.info('{begin}{msg:>{fill}}{end:>{e_fill}}'.format(begin=fence,
                                                                 msg=msg,
                                                                 fill=(width + len(msg)) // 2,
                                                                 end=fence,
                                                                 e_fill=(width - len(msg) - 1) // 2))
        logger.info(fence * width)

    message('Welcome to Eskapade!')

    # check for batch mode
    if settings.get('batchMode'):
        # set non-interactive Matplotlib backend before plotting tools are imported
        eskapade.utils.set_matplotlib_backend(batch=True, silent=False)

    # execute configuration macro, this sets up the order of the chains and links.
    if not settings['macro']:
        raise RuntimeError('macro is not set')
    process_manager.execute_macro(settings['macro'])

    # check analysis name
    if not settings['analysisName']:
        raise RuntimeError('analysis name is not set')

    # standard execution from now on
    status = process_manager.run()

    message('Leaving Eskapade. Bye!')

    return status



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk204_apply_query_to_pandas_df

Created: 2017/02/23

Description:
    Illustrates link that applies basic queries to pandas dataframe
    See for more information pandas documentation:

    http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.query.html

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from numpy.random import randn
from pandas import DataFrame

from eskapade import ConfigObject, DataStore, Chain
from eskapade import core_ops, analysis
from eskapade import process_manager
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk204_apply_query_to_pandas_df')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk204_apply_query_to_pandas_df'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# generate a dummy dataframe and add to datastore
# to this dataset selections are applied below, during link execution.

# NB: realize that, normally, such a dataframe is read or constructed on the fly
# during link execution.


df = DataFrame(randn(100, 3), columns=list('abc'))

ds = process_manager.service(DataStore)
ds['incoming_records'] = df

#########################################################################################
# --- Here we apply example selections to a dataframe picked up from the datastore.

data_prep = Chain('DataPrep')

# query_set = seletions that are applies to incoming_records
# after selections, only keep column in select_columns ('a', 'c')
link = analysis.ApplySelectionToDf(read_key='incoming_records',
                                   store_key='outgoing_records',
                                   query_set=['a>0', 'c<b'],
                                   select_columns=['a', 'c'])
# Any other kwargs given to ApplySelectionToDf are passed on the the
# pandas query() function.
link.logger.log_level = LogLevel.DEBUG
data_prep.add(link)

link = core_ops.DsObjectDeleter()
link.deletion_keys = ['incoming_records']
data_prep.add(link)

link = core_ops.PrintDs()
link.keys = ['n_outgoing_records', 'outgoing_records']
data_prep.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk204_apply_query_to_pandas_df')



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk301_dfsummary_plotter

Created: 2017/02/23

Description:
    Macro shows how to plot the content of a dataframe in a nice summary
    pdf file.
    (Example similar to tutorial_1)

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import analysis, visualization
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk301_dfsummary_plotter.')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk301_dfsummary_plotter'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

msg = r"""

The plots and latex files produced by link df_summary can be found in dir:
{path}
"""
logger.info(msg, path=settings['resultsDir'] + '/' + settings['analysisName'] + '/data/v0/report/')

COLUMNS = ['var_a', 'var_b', 'var_c']
SIZE = 10000
VAR_LABELS = dict(var_a='Variable A', var_b='Variable B', var_c='Variable C')
VAR_UNITS = dict(var_b='m/s')
GEN_CONF = dict(var_b=dict(mean=42., std=2.), var_c=dict(mean=42, std=2, dtype=int))

#########################################################################################
# --- now set up the chains and links based on configuration flags

data = Chain('Data')
# add data-generator link to "Data" chain
generator = analysis.BasicGenerator(name='Generate_data',
                                    key='data',
                                    columns=COLUMNS,
                                    size=SIZE,
                                    gen_config=GEN_CONF)
data.add(generator)

# add data-frame summary link to "Summary" chain
# can provide labels and units for the variables in the dataset
summary = Chain('Summary')
summarizer = visualization.DfSummary(name='Create_stats_overview',
                                     read_key=generator.key,
                                     var_labels=VAR_LABELS,
                                     var_units=VAR_UNITS)
summary.add(summarizer)

#########################################################################################

logger.debug('Done parsing configuration file esk301_dfsummary_plotter')



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk105_datastore_pickling

Created: 2017/02/20

Description:
    Macro serves as input to other three esk105 example macros.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import core_ops
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk105_datastore_pickling.')

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk105_datastore_pickling'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

msg = r"""

The setup consists of three simple chains that add progressively more information to the datastore.
In the examples the datastore gets persisted after the execution of each chain, and can be picked
up again as input for the next chain.

- The pickled datastore(s) can be found in the data directory:
{data_path}

- The pickled configuration object(s) and backed-up configuration file can be found in:
{conf_path}
"""
logger.info(msg, data_path=settings['resultsDir'] + '/' + settings['analysisName'] + '/data/v0/',
            conf_path=settings['resultsDir'] + '/' + settings['analysisName'] + '/config/v0/')

# dummy information used in this macro, added to each chain below.
f = {'hello': 'world', 'v': [3, 1, 4, 1, 5], 'n_favorite': 7}
g = {'a': 1, 'b': 2, 'c': 3}
h = [2, 7]

#########################################################################################
# --- now set up the chains and links based on configuration flags

#########
# chain 1
ch = Chain('chain1')

# the link ToDsDict adds objects to the datastore at link execution.
link = core_ops.ToDsDict(name='intods_1')
link.store_key = 'f'
link.obj = f
ch.add(link)

# print contents of datastore
link = core_ops.PrintDs()
ch.add(link)

#########
# chain 2
ch = Chain('chain2')

# the link AssertInDs checks the presence
# of certain objects in the datastore
link = core_ops.AssertInDs()
link.keySet = ['f']
ch.add(link)

# the link ToDsDict adds objects to the datastore at link execution.
link = core_ops.ToDsDict(name='intods_2')
link.store_key = 'g'
link.obj = g
ch.add(link)

link = core_ops.PrintDs()
ch.add(link)

#########
# chain 3
ch = Chain('chain3')

# the link AssertInDs checks the presence
# of certain objects in the datastore
link = core_ops.AssertInDs()
link.keySet = ['f', 'g']
ch.add(link)

# the link ToDsDict adds objects to the datastore at link execution.
link = core_ops.ToDsDict(name='intods_3')
link.store_key = 'h'
link.obj = h
ch.add(link)

link = core_ops.PrintDs()
ch.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk105_datastore_pickling.')



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/08/23

Description:

   Logging module for Eskapade.

   This module is thin layer on top of the Python logging module
   that provides a simple interface to the user and PEP-3101
   (Advanced String Formatting) with named placeholders.

   Sources of inspiration for this module are:

       - The Twisted Logger interface.
       - The Python logging cookbook and how to.

Authors:
   KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import inspect
import logging
import sys
import time
from enum import IntEnum, unique
from typing import Union, Any

import pendulum


# Protected. Should not be exported.
class _Message(object):
    __slots__ = 'fmt', 'kwargs'

    def __init__(self, fmt: str, kwargs) -> None:
        # WARNING is seven characters.
        self.fmt = fmt
        self.kwargs = kwargs

    def __str__(self) -> str:
        return self.fmt.format(**self.kwargs)


@unique
class LogLevel(IntEnum):
    """Logging level integer enumeration class.

    The enumerations are:

    * NOTSET  ( 0)
    * DEBUG   (10)
    * INFO    (20)
    * WARNING (30)
    * ERROR   (40)
    * FATAL   (50)

    They have the same meaning and value as the logging levels in the
    Python logging module.
    """

    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    FATAL = logging.FATAL

    def __str__(self) -> str:
        return self.name


# Publishers.
class LogPublisher(logging.getLoggerClass()):
    """Logging publisher that listens for log events."""

    def __init__(self, name: str = '', level: Union[int, LogLevel] = LogLevel.INFO):
        """Initialize the publisher."""
        super().__init__(name, level)

    def log(self, _level: LogLevel, _msg: str, *args, **kwargs):
        """Log a message with log info."""
        if self.isEnabledFor(_level):
            exc_info = kwargs.pop('exc_info', False)
            extra = kwargs.pop('extra', None)
            stack_info = kwargs.pop('stack_info', False)

            human_time = pendulum.utcfromtimestamp(kwargs.get('log_time', time.time()))
            kwargs['log_time'] = str(human_time)
            # We need LogLevel(level) to convert logging level from logging to LogLevel
            # and get its str representation.
            kwargs['log_level'] = str(kwargs.get('log_level', LogLevel(_level)))
            kwargs['log_namespace'] = kwargs.get('log_namespace', 'publisher')

            _msg = '{log_time!s} [{log_namespace!s}#{log_level!s}] ' + _msg
            self._log(_level,
                      _Message(_msg, kwargs),
                      args,
                      exc_info=exc_info,
                      extra=extra,
                      stack_info=stack_info)

    def add_handler(self, handler: logging.StreamHandler) -> None:
        """Add a log record handler.

        :param handler: The log record handler.
        """
        self.addHandler(handler)

    def remove_handler(self, handler: logging.StreamHandler):
        """Remove a log record handler.

        :param handler: The log record handler.
        """
        self.removeHandler(handler)

    def event(self, event: dict) -> None:
        """Log a logging event.

        This method is called by event emitters, e.g. :class:`Logger`.

        :param event: The event to log.
        :type event: dict
        """
        self.log(event['log_level'], event.pop('log_msg'), **event)

    @property
    def log_level(self):
        """Get logging level.

        :return: The logging level.
        :rtype: LogLevel
        """
        return LogLevel(self.level)

    @log_level.setter
    def log_level(self, level: LogLevel):
        """Set the logging level.

        :param level: The logging level to set/use.
        :type level: LogLevel
        """
        if level in LogLevel:
            self.setLevel(level)
            return

        raise ValueError('Unknown logging level {level!s}'.format(level=level))


# Set LogPublisher to be the logger class.
logging.setLoggerClass(LogPublisher)

# The global application publisher. There should be only one.
global_log_publisher = logging.getLogger('eskapade')


# Handlers
class ConsoleHandler(logging.StreamHandler):
    """A stream handler that sends log messages with levels up to and including WARNING to stdout."""

    def __init__(self):
        """Initialize the ConsoleHandler object."""
        super().__init__(sys.stdout)
        self.max_log_level = LogLevel.WARNING

    def filter(self, record):
        """Determine if a record should be sent to stdout."""
        return record.levelno <= self.max_log_level


class ConsoleErrHandler(logging.StreamHandler):
    """A stream handler that sends log messages with level ERROR and above to stderr."""

    def __init__(self):
        """Initialize the ConsoleHandler object."""
        super().__init__(sys.stderr)
        self.setLevel(LogLevel.ERROR)


class Logger(object):
    """A logger that emits log messages to an observer.

    The logger can be instantiated as a module or class attribute, e.g.

    >>> logger = Logger()
    >>> logger.info("I'm a module logger attribute.")
    >>>
    >>> class Point(object):
    >>>     logger = Logger()
    >>>
    >>>     def __init__(self, x = 0.0, y = 0.0):
    >>>         Point.logger.debug('Initializing {point} with x = {x}  y = {y}', point=Point, x=x, y=y)
    >>>         self._x = x
    >>>         self._y = y
    >>>
    >>>     @property
    >>>     def x(self):
    >>>         self.logger.debug('Getting property x = {point._x}', point=self)
    >>>         return self._x
    >>>
    >>>     @x.setter
    >>>     def x(self, x):
    >>>         self.logger.debug('Setting property y = {point._x}', point=self)
    >>>         self._x = x
    >>>
    >>>     @property
    >>>     def y(self):
    >>>        self.logger.debug('Getting property y = {point._y}', point=self)
    >>>        return self._y
    >>>
    >>>     @y.setter
    >>>     def y(self, y):
    >>>         self.logger.debug('Setting property y = {point._y}', point=self)
    >>>         self._y = y
    >>>
    >>> a_point = Point(1, 2)
    >>>
    >>> logger.info('p_x = {point.x} p_y = {point.y}', point=a_point)
    >>> logger.log_level = LogLevel.DEBUG
    >>> logger.info('p_x = {point.x} p_y = {point.y}', point=a_point)

    The logger uses PEP-3101 (Advanced String Formatting) with named placeholders,
    see <https://www.python.org/dev/peps/pep-3101/>
    and <https://pyformat.info/> for more details and examples.

    Furthermore, logging events are only formatted and evaluated for logging
    levels that are enabled. So, there's no need to check the logging level
    before logging. It's also efficient.
    """

    @staticmethod
    def __namespace_from_calling_context() -> str:
        """Private static method to derive namespace/calling context.

        :return: The calling frame context.
        :rtype: str
        """
        # We use the inspect module to get and traverse the stack.
        # Index 0 - Me.
        # Index 1 - The guy calling me. In this case __init__.
        # Index 2 - The guy calling/instantiating Logger.
        calling_frame_context = inspect.stack()[2].frame.f_globals["__name__"]

        return calling_frame_context

    def __init__(self, name: str = None, source: str = None, observer: Union[LogPublisher, logging.Logger] = None):
        """Initialize logger instance."""
        name = name or self.__namespace_from_calling_context()

        self.name = name
        self.source = source
        self.observer = observer or global_log_publisher

    def __get__(self, instance: Any, instance_type=None):
        """Create or get a logger.

        When used as a descriptor, i.e.::

            # File: a_module.py
            class Klass(object):
                 logger = Logger()

                 def hello(self):
                     self.logger.info('Hello')

        The logger's name will be set to the name of the class it's declared on. In the above example
        the name of the logger is :class:`a_module.Klass`. Also, :attr:`Klass.logger.source` would be
        :class:`a_module.Klass` and :attr:`Klass().logger.source` would be an instance of
        :class:`a_module.Klass`.

        Note that :func:`logging.getLogger` will either create or get a logger with said name.
        """
        source = instance or instance_type
        paths = [instance_type.__module__, instance_type.__name__]
        if not instance_type.__module__.startswith(global_log_publisher.name + '.'):
            paths.insert(0, global_log_publisher.name)
        name = '.'.join(paths)
        return self.__class__(name=name,
                              source=source,
                              observer=logging.getLogger(name))

    def __set__(self, instance: Any, value):
        raise AttributeError('Cannot redefine logger for {instance!s}!'.format(instance=instance))

    def __repr__(self):
        return '<{klass!s} ' \
               'name={namespace!r} ' \
               'source={source!r} ' \
               'observer={observer!r} ' \
               'level={observer_level!r} ' \
               'id=\'{id!r}\'>'.format(klass=self.__class__.__name__,
                                       namespace=self.name,
                                       source=self.source,
                                       observer=self.observer,
                                       observer_level=self.observer.log_level,
                                       id=id(self), )

    def __log(self, log_level: LogLevel, fmt: str = '', **kwargs):
        # In Python 3.6 the order of kwargs is preserved, see PEP 468.
        # So, we are not going to bother to fix it here.
        event = kwargs
        if not kwargs:
            fmt = fmt.replace('{', '{{').replace('}', '}}')
        fmt = fmt or ' '.join(k + ' = {' + k + '}' for k in event.keys())
        event.update(log_time=time.time(),
                     log_level=log_level,
                     log_namespace=self.name,
                     log_source=self.source,
                     log_msg=fmt)
        self.observer.event(event)

    def debug(self, fmt: str = '', **kwargs) -> None:
        """Emit a log event at the DEBUG level.

        DEBUG log events are only formatted and evaluated if DEBUG is enabled.

        :param fmt: The DEBUG event string to be formatted.
        :type fmt: str
        :param kwargs: The placeholder keywords and their values.
        """
        self.__log(LogLevel.DEBUG, fmt, **kwargs)

    def info(self, fmt: str = '', **kwargs) -> None:
        """Emit a log event at the INFO level.

        INFO log events are only formatted and evaluated if INFO is enabled.

        :param fmt: The INFO event string to be formatted.
        :type fmt: str
        :param kwargs: The placeholder keywords and their values.
        """
        self.__log(LogLevel.INFO, fmt, **kwargs)

    def warning(self, fmt: str = '', **kwargs) -> None:
        """Emit a log event at the WARNING level.

        WARNING log events are only formatted and evaluated if WARNING is enabled.

        :param fmt: The WARNING event string to be formatted.
        :type fmt: str
        :param kwargs: The placeholder keywords and their values.
        """
        self.__log(LogLevel.WARNING, fmt, **kwargs)

    def error(self, fmt: str = '', **kwargs) -> None:
        """Emit a log event at the ERROR level.

        ERROR log events are only formatted and evaluated if ERROR is enabled.

        :param fmt: The ERROR event string to be formatted.
        :type fmt: str
        :param kwargs: The placeholder keywords and their values.
        """
        self.__log(LogLevel.ERROR, fmt, **kwargs)

    def fatal(self, fmt: str = '', **kwargs) -> None:
        """Emit a log event at the FATAL level.

        FATAL log events are only formatted and evaluated if FATAL is enabled.

        :param fmt: The ERROR event string to be formatted.
        :type fmt: str
        :param kwargs: The placeholder keywords and their values.
        """
        self.__log(LogLevel.FATAL, fmt, **kwargs)

    @property
    def log_level(self) -> LogLevel:
        """Get the logging level of the logger.

        :return: The logging level.
        :rtype: LogLevel
        """
        return self.observer.log_level

    @log_level.setter
    def log_level(self, level: LogLevel):
        """Set the logging level of the logger.

        :param level:
        :type level: LogLevel
        """
        self.observer.log_level = level



"""Project: Eskapade - A python-based package for data analysis.

Created: 2016/11/08

Description:
    Utility functions to collect Eskapade python modules
    e.g. functions to get correct Eskapade file paths and env variables

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import os
import sys

import matplotlib

from eskapade.logger import Logger

ENV_VARS = dict(spark_args='PYSPARK_SUBMIT_ARGS',
                docker='DE_DOCKER', display='DISPLAY')
ARCHIVE_FILE = 'es_python_modules.egg'

logger = Logger()


def set_matplotlib_backend(backend=None, batch=None, silent=True):
    """Set Matplotlib backend.

    :param str backend: backend to set
    :param bool batch: require backend to be non-interactive
    :param bool silent: do not raise exception if backend cannot be set
    :raises: RuntimeError
    """
    # determine if batch mode is required
    display = get_env_var('display')
    run_batch = bool(batch) or display is None or not display.startswith(':') or not display[1].isdigit()
    if run_batch:
        matplotlib.interactive(False)

    # check if batch mode can be set if it is requested
    if not batch and batch is not None and run_batch:
        if not silent:
            raise RuntimeError('Interactive Matplotlib mode requested, but no display found.')
        logger.warning('Matplotlib cannot be used interactively; no display found.')

    # get Matplotlib backends
    curr_backend = matplotlib.get_backend().lower()
    ni_backends = [nib.lower() for nib in matplotlib.rcsetup.non_interactive_bk]

    # determine backend to be set
    if not backend:
        # try to use current backend
        backend = curr_backend if not run_batch or curr_backend in ni_backends else ni_backends[0]
    backend = str(backend).lower()

    # check if backend is compatible with mode
    if run_batch and backend not in ni_backends:
        if not silent:
            raise RuntimeError('Non-interactive Matplotlib backend required, but "{!s}" requested.'.format(backend))
        logger.warning(
            'Set Matplotlib backend to "{actual}"; non-interactive backend required, but "{requested}" requested.',
            actual=ni_backends[0], requested=backend)
        backend = ni_backends[0]

    # check if backend has to change
    if backend == curr_backend:
        return

    # check if backend can still be set
    if 'matplotlib.pyplot' in sys.modules:
        if not silent:
            raise RuntimeError('cannot set Matplotlib backend: pyplot module already loaded.')
        return

    # set backend
    matplotlib.use(backend)


def get_env_var(key):
    """Retrieve Eskapade-specific environment variables.

    :param str key: Eskapade-specific key to variable
    :returns: environment variable value
    :rtype: str
    """
    var_name = ENV_VARS[key]
    return os.environ.get(var_name)


def collect_python_modules():
    """Collect Eskapade Python modules."""
    import pathlib
    from pkg_resources import resource_filename
    from zipfile import PyZipFile

    import eskapade

    package_dir = resource_filename(eskapade.__name__, '')
    lib_path = pathlib.Path(package_dir).joinpath('lib')
    lib_path.mkdir(exist_ok=True)
    archive_path = str(lib_path.joinpath(ARCHIVE_FILE))

    archive_file = PyZipFile(archive_path, 'w')
    logger.info('Adding Python modules to egg archive {path}.'.format(path=archive_path))
    archive_file.writepy(package_dir)
    archive_file.close()
    return archive_path



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/04/11

Description:
    Utilities for Eskapade run

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import argparse

from eskapade.core.definitions import USER_OPTS
from eskapade.core.definitions import USER_OPTS_KWARGS
from eskapade.core.definitions import USER_OPTS_SHORT


def create_arg_parser():
    """Create parser for user arguments.

    An argparse parser is created and returned, ready to parse
    arguments specified by the user on the command line.

    :returns: argparse.ArgumentParser
    """
    # create parser and add arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('config_files', nargs='+', metavar='CONFIG_FILE', help='configuration file to execute')
    for sec_keys in USER_OPTS.values():
        for opt_key in sec_keys:
            args = ['--{}'.format(opt_key).replace('_', '-')]
            if opt_key in USER_OPTS_SHORT:
                args.append('-{}'.format(USER_OPTS_SHORT[opt_key]))
            parser.add_argument(*args, **USER_OPTS_KWARGS.get(opt_key, {}))

    return parser



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/08/23

Description:
    Collection of helper functions to get fixtures, i.e. test data,
    ROOT/RooFit libs, and tutorials. These are mostly used by the
    (integration) tests.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import pathlib
import sys

from pkg_resources import resource_filename

import eskapade

# Fixtures
_FIXTURE = {_.name: _ for _ in pathlib.Path(resource_filename(eskapade.__name__, 'data')).glob('*')}

# C/C++ libraries that are shipped with eskapade.
if sys.platform == 'darwin':
    _LIBS = {_.name: _ for _ in pathlib.Path(resource_filename(eskapade.__name__, 'lib')).glob('*.dylib')}
else:
    _LIBS = {_.name: _ for _ in pathlib.Path(resource_filename(eskapade.__name__, 'lib')).glob('*.so')}

# Tutorials that are shipped with eskapade.
_TUTORIALS = {_.name if _.parent.name == 'tutorials' else _.parent.name + '/' + _.name:
              _ for _ in pathlib.Path(resource_filename(eskapade.__name__, 'tutorials')).glob('**/*.py')}

# Templates that are shipped with eskapade.
_TEMPLATES = {_.name: _ for _ in pathlib.Path(resource_filename(eskapade.__name__, 'templates')).glob('*')}

# Configs that are shipped with eskapade.
_CONFIGS = {_.name: _ for _ in pathlib.Path(resource_filename(eskapade.__name__, 'config')).glob('**/*')}

# Resource types
_RESOURCES = {
    'fixture': _FIXTURE,
    'library': _LIBS,
    'tutorial': _TUTORIALS,
    'template': _TEMPLATES,
    'config': _CONFIGS
}


def _resource(resource_type, name: str) -> str:
    """Return the full path filename of a resource.

    :param str resource_type: The type of the resource.
    :param str  name: The name of the resource.
    :returns: The full path filename of the fixture data set.
    :rtype: str
    :raises FileNotFoundError: If the resource cannot be found.
    """
    full_path = _RESOURCES[resource_type].get(name, None)

    if full_path and full_path.exists():
        return str(full_path)

    raise FileNotFoundError('Could not find {resource_type} "{name!s}"! Does it exist?'
                            .format(resource_type=resource_type, name=name))


def fixture(name: str) -> str:
    """Return the full path filename of a fixture data set.

    :param str name: The name of the fixture.
    :returns: The full path filename of the fixture data set.
    :rtype: str
    :raises FileNotFoundError: If the fixture cannot be found.
    """
    return _resource('fixture', name)


def lib(name: str) -> str:
    """Return the full path filename of a library.

    :param str name: The name of the library.
    :returns: The full path filename of the library.
    :rtype: str
    :raises FileNotFoundError: If the library cannot be found.
    """
    return _resource('library', name)


def tutorial(name: str) -> str:
    """Return the full path filename of a tutorial.

    :param str name: The name of the tutorial.
    :returns: The full path filename of the tutorial.
    :rtype: str
    :raises FileNotFoundError: If the tutorial cannot be found.
    """
    return _resource('tutorial', name)


def template(name: str) -> str:
    """Return the full path filename of a tutorial.

    :param str name: The name of the template.
    :returns: The full path filename of the tutorial.
    :rtype: str
    :raises FileNotFoundError: If the template cannot be found.
    """
    return _resource('template', name)


def config(name: str) -> str:
    """Return the absolute path of a config.

    :param str name: The name of the config.
    :returns: The absolute path of the config.
    :raises FileNotFoundError: If the config cannot be found.
    """
    return _resource('config', name)



"""Project: Eskapade - A python-based package for data analysis.

Class: RepeatChain

Created: 2016/11/08

Description:
    Algorithm that sends "repeat this chain" signal to processManager,
    until ready.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import process_manager, ConfigObject, Link, StatusCode


class RepeatChain(Link):
    """Algorithm that sends signal to processManager to repeat the current chain."""

    def __init__(self, **kwargs):
        """Link that sends signal to processManager to repeat the current chain.

        Sents a RepeatChain deenums.StatusCode signal.

        :param str name: name of link
        :param list listen_to: repeat this chain if given key is present in ConfigObject and set to true.
            E.g. this key is set by readtods link when looping over files.
        :param int maxcount: repeat this chain until max count has been reacher. Default is -1 (off).
        """
        # initialize Link
        Link.__init__(self, kwargs.pop('name', 'RepeatChain'))

        # process keyword arguments
        self._process_kwargs(kwargs, maxcount=-1, listen_to=[])
        self.check_extra_kwargs(kwargs)

        self._counter = 0

    def initialize(self):
        """Initialize the link."""
        if isinstance(self.listen_to, list):
            pass
        elif isinstance(self.listen_to, str) and self.listen_to:
            self.listen_to = [self.listen_to]
        else:
            raise Exception('listen_to key of incorrect type.')

        return StatusCode.Success

    def execute(self):
        """Execute the link."""
        settings = process_manager.service(ConfigObject)

        # search for listen_to key in ConfigObject. if present and true, send signal to repeat current chain.
        for l in self.listen_to:
            if l in settings and settings[l]:
                self.logger.debug('The repeater count is: {n:d}.', n=self._counter)
                self._counter += 1
                return StatusCode.RepeatChain

        # repeat this chain until counter reaches specified maxcount value..
        while self._counter < self.maxcount:
            self.logger.debug('The repeater count is: {n:d}. Max count is: {max:d}', n=self._counter, max=self.maxcount)
            self._counter += 1
            return StatusCode.RepeatChain

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Class: DfMerger

Created: 2016/11/08

Description:
    Algorithm to Merges two pandas DataFrames

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy

import pandas as pd

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class DfMerger(Link):
    """Merges two pandas dataframes."""

    def __init__(self, **kwargs):
        """Initialize link instance.

        Store the configuration of the link.

        :param str name: name of link
        :param str input_collection1: datastore key of the first pandas.DataFrame to merge
        :param str input_collection2: datastore key of the second pandas.DataFrame to merge
        :param str output_collection: datastore key of the merged output pandas.DataFrame
        :param str how: merge modus. See pandas documentation.
        :param list on: column names. See pandas documentation.
        :param list columns1: column names of the first pandas.DataFrame. Only these columns are included in the merge.
            If not set, use all columns.
        :param list columns2: column names of the second pandas.DataFrame. Only these columns are included in the merge.
            If not set, use all columns.
        :param bool remove_duplicate_cols2: if True duplicate columns will be taken out before the merge (default=True)
        :param kwargs: all other key word arguments are passed on to the pandas merge function.
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'DfMerger'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             input_collection1=None,
                             input_collection2=None,
                             output_collection='',
                             how='inner',
                             on=['record_id'],
                             columns1=[],
                             columns2=[],
                             remove_duplicate_cols2=True)

        # pass on remaining kwargs to pandas reader
        self.kwargs = copy.deepcopy(kwargs)

    def initialize(self):
        """Perform basic checks on provided attributes."""
        if isinstance(self.on, str):
            self.on = [self.on]
        elif not isinstance(self.on, list):
            raise Exception('on is not a list of strings.')

        assert len(self.on) > 0, 'Not specified on which keys to merge.'
        assert (self.how == 'inner' or self.how == 'outer' or self.how == 'left' or self.how == 'right'), \
            'How to merge not specified correctly.'
        assert len(self.output_collection), 'output_collection not specified.'

        # add back on to kwargs, so it's picked up by pandas.
        if self.on is not None:
            self.kwargs['on'] = self.on
        if self.how is not None:
            self.kwargs['how'] = self.how

        self.logger.info('kwargs passed on to pandas merge function are: {kwargs}.', kwargs=self.kwargs)

        return StatusCode.Success

    def execute(self):
        """Perform merging of input dataframes."""
        ds = process_manager.service(DataStore)

        # Perform basic checks on input collections and provided attributes.

        # datasets okay?
        assert self.input_collection1 in ds, 'Key {} not in DataStore.'.format(self.input_collection1)
        assert self.input_collection2 in ds, 'Key {} not in DataStore.'.format(self.input_collection2)
        df1 = ds[self.input_collection1]
        df2 = ds[self.input_collection2]
        assert isinstance(df1, pd.DataFrame), 'Item {} is not a DataFrame.'.format(self.input_collection1)
        assert isinstance(df2, pd.DataFrame), 'Item {} is not a DataFrame.'.format(self.input_collection2)

        # check provided columns
        if len(self.columns1) == 0:
            self.columns1 = df1.columns
        for col in self.columns1:
            assert col in df1.columns, 'Column <{}> not in input_collection1'.format(col)
        if len(self.columns2) == 0:
            self.columns2 = df2.columns
        for col in self.columns2:
            assert col in df2.columns, 'Column <{}> not in input_collection2'.format(col)
        for col in self.on:
            assert col in self.columns1, 'Key <{}> not in selected columns input_collection1.'.format(col)
            assert col in self.columns2, 'Key <{}> not in selected columns input_collection2.'.format(col)
        if self.remove_duplicate_cols2:
            self.columns2 = [c for c in self.columns2 if c not in self.columns1]
            self.columns2 += self.on

        # then do the merging.
        dfout = pd.merge(df1[self.columns1], df2[self.columns2], **self.kwargs)

        # storage
        ds[self.output_collection] = dfout
        ds['n_' + self.output_collection] = len(dfout.index)

        self.logger.info('Put dataframe <{df}> with length <{length:d}> in data store.',
                         df=self.output_collection, length=len(dfout.index))

        return StatusCode.Success



from eskapade.visualization.links.correlation_summary import CorrelationSummary
from eskapade.visualization.links.df_boxplot import DfBoxplot
from eskapade.visualization.links.df_summary import DfSummary

__all__ = ['CorrelationSummary', 'DfBoxplot', 'DfSummary']



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/02/27

Description:
    Definitions used in Eskapade runs:
    * logging levels
    * return-status codes
    * default configuration variables
    * user options

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import ast
import collections
import os
from enum import IntEnum, unique

from pkg_resources import resource_filename

from eskapade.logger import LogLevel


@unique
class StatusCode(IntEnum):
    """Return status code enumeration class.

    A StatusCode should be returned by the initialize, execute,
    and finalize methods of links, chains, and the process manager.

    The enumerations are:

    * Undefined (-1): Default status.
    * Success (0 == EX_OK / EXIT_SUCCESS): All OK, i.e. there were no errors.
    * RepeatChain (1): Repeat execution of this chain.
    * SkipChain (2): Skip this chain: initialize, execute, and finalize.
    * BreakChain (3): Skip the further execution of this this, but do perform finalize.
    * Recoverable (4): Not OK, but can continue, i.e. there was an error, but the
      application can recover from it.
    * Failure (5): An error occurred and the application cannot recover from it.
      In this case the application should just quit.
    """

    Undefined = -1  # type: int
    Success = 0  # type: int
    RepeatChain = 1  # type: int
    SkipChain = 2  # type: int
    BreakChain = 3  # type: int
    Recoverable = 4  # type: int
    Failure = 5  # type: int

    def __str__(self) -> str:
        """Get string representation of :class:`StatusCode`.

        :return: String representation of :class:`StatusCode`.
        :rtype: str
        """
        return self.name

    def is_undefined(self) -> bool:
        """Check if status is `StatusCode.Undefined`.

        :return: True when `StatusCode.Undefined`, False otherwise.
        :rtype: bool
        """
        return StatusCode.Undefined == self

    def is_success(self) -> bool:
        """Check if status is `StatusCode.Success`.

        :return: True when `StatusCode.Success`, False otherwise.
        :rtype: bool
        """
        return StatusCode.Success == self

    def is_repeat_chain(self) -> bool:
        """Check if status is `StatusCode.RepeatChain`.

        :return: True when `StatusCode.RepeatChain`, False otherwise.
        :rtype: bool
        """
        return StatusCode.RepeatChain == self

    def is_skip_chain(self) -> bool:
        """Check if status is `StatusCode.SkipChain`.

        :return: True when `StatusCode.SkipChain`, False otherwise.
        :rtype: bool
        """
        return StatusCode.SkipChain == self

    def is_break_chain(self) -> bool:
        """Check if status is `StatusCode.BreakChain`.

        :return: True when `StatusCode.BreakChain`, False otherwise.
        :rtype: bool
        """
        return StatusCode.BreakChain == self

    def is_recoverable(self) -> bool:
        """Check if status is `StatusCode.Recoverable`.

        :return: True when `StatusCode.Recoverable`, False otherwise.
        :rtype: bool
        """
        return StatusCode.Recoverable == self

    def is_failure(self) -> bool:
        """Check if status is `StatusCode.Failure`.

        :return: True when `StatusCode.Failure`, False otherwise.
        :rtype: bool
        """
        return StatusCode.Failure == self


class RandomSeeds:
    """Container for seeds of random generators.

    Seeds are stored as key-value pairs and are accessed with getitem and
    setitem methods.  A default seed can be accessed with the key "default".
    The default seed is also returned if no seed is set for the specified
    key.

    >>> import numpy as np
    >>> seeds = RandomSeeds(default=999, foo=42, bar=13)
    >>> seeds['NumPy'] = 100
    >>> np.random.seed(seeds['NumPy'])
    >>> print(seeds['nosuchseed'])
    999
    """

    def __init__(self, **kwargs):
        """Initialize an instance.

        Values of the specified keyword arguments must be integers, which are
        set as seed values for the corresponding key.
        """
        # initialize attributes
        self._seeds = {}
        self._default = 1

        # set specified seeds
        for key, seed in kwargs.items():
            self[key] = seed

    def __getitem__(self, key):
        """Return seed for specified lowercase-string key."""
        return self._seeds.get(str(key).strip().lower(), self._default)

    def __setitem__(self, key, seed):
        """Set integer seed for specified lowercase-string key."""
        # parse key and seed
        key = str(key).strip().lower()
        try:
            seed = int(seed)
        except Exception:
            raise TypeError('specified seed for key "{0:s}" is not an integer: "{1!s}"'.format(key, seed))

        # check if this is the default key
        if key == 'default':
            self._default = seed
        else:
            self._seeds[key] = seed

    def __str__(self):
        seed_str = ', '.join('{0:s}: {1:d}'.format(*kv) for kv in self._seeds.items())
        return '{{default: {0:d} | {1:s}}}'.format(self._default, seed_str)


# configuration variables
CONFIG_VARS = collections.OrderedDict()

CONFIG_VARS['run'] = ['analysisName',
                      'version',
                      'macro',
                      'batchMode',
                      'interactive',
                      'logLevel',
                      'logFormat',
                      'doCodeProfiling', ]

CONFIG_VARS['chains'] = ['beginWithChain',
                         'endWithChain',
                         'storeResultsEachChain',
                         'storeResultsOneChain',
                         'doNotStoreResults', ]

CONFIG_VARS['file_io'] = ['esRoot',
                          'resultsDir',
                          'dataDir',
                          'macrosDir',
                          'templatesDir',
                          'configDir', ]

CONFIG_VARS['config'] = ['sparkCfgFile', ]

CONFIG_VARS['db_io'] = ['all_mongo_collections', ]

CONFIG_VARS['rand_gen'] = ['seeds', ]

CONFIG_TYPES = dict(version=int,
                    batchMode=bool,
                    interactive=bool,
                    storeResultsEachChain=bool,
                    doNotStoreResults=bool,
                    all_mongo_collections=list, )

CONFIG_DEFAULTS = dict(version=0,
                       batchMode=True,
                       interactive=False,
                       logLevel=LogLevel.INFO,
                       logFormat='%(asctime)s %(levelname)s [%(module)s]: %(message)s',
                       doCodeProfiling=None,
                       storeResultsEachChain=False,
                       doNotStoreResults=False,
                       esRoot=os.getcwd() + '/',
                       resultsDir=os.getcwd() + '/results/',
                       dataDir=os.getcwd() + '/data/',
                       macrosDir=os.getcwd() + '/macros/',
                       templatesDir=resource_filename('eskapade', 'templates') + '/',
                       configDir=os.getcwd() + '/config/',
                       sparkCfgFile='spark.cfg',
                       seeds=RandomSeeds(), )

# user options in command-line arguments
USER_OPTS = collections.OrderedDict()
USER_OPTS['run'] = ['analysis_name',
                    'analysis_version',
                    'batch_mode',
                    'interactive',
                    'log_level',
                    'log_format',
                    'unpickle_config',
                    'profile',
                    'conf_var', ]

USER_OPTS['chains'] = ['begin_with',
                       'end_with',
                       'single_chain',
                       'store_all',
                       'store_one',
                       'store_none', ]

USER_OPTS['file_io'] = ['results_dir',
                        'data_dir',
                        'macros_dir',
                        'templates_dir', ]

USER_OPTS['config'] = ['spark_cfg_file', ]

USER_OPTS['rand_gen'] = ['seed', ]

USER_OPTS_SHORT = dict(analysis_name='n',
                       analysis_version='v',
                       interactive='i',
                       log_level='L',
                       conf_var='c',
                       begin_with='b',
                       end_with='e',
                       single_chain='s', )

USER_OPTS_KWARGS = dict(analysis_name=dict(help='set name of analysis in run',
                                           metavar='NAME'),
                        analysis_version=dict(help='set version of analysis version in run',
                                              type=int,
                                              metavar='VERSION'),
                        batch_mode=dict(help='run in batch mode (no X Windows)',
                                        action='store_true'),
                        interactive=dict(help='start Python shell after run',
                                         action='store_true'),
                        log_level=dict(help='set logging level',
                                       choices=['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'FATAL'],
                                       metavar='{NOTSET,DEBUG,INFO,WARNING,ERROR,FATAL}'),
                        log_format=dict(help='set log-message format',
                                        metavar='FORMAT'),
                        unpickle_config=dict(help='interpret first CONFIG_FILE as path to pickled settings',
                                             action='store_true'),
                        profile=dict(help='run Python profiler, sort output by specified column',
                                     choices=['stdname', 'nfl', 'pcalls', 'file', 'calls', 'time', 'line',
                                              'cumulative', 'module', 'name'],
                                     metavar='{stdname,nfl,pcalls,file,calls,time,line,cumulative,module,name}'),
                        conf_var=dict(help='set configuration variable',
                                      action='append',
                                      metavar='KEY=VALUE'),
                        begin_with=dict(help='begin execution with chain CHAIN_NAME',
                                        metavar='CHAIN_NAME'),
                        end_with=dict(help='end execution with chain CHAIN_NAME',
                                      metavar='CHAIN_NAME'),
                        single_chain=dict(help='only execute chain CHAIN_NAME',
                                          metavar='CHAIN_NAME'),
                        store_all=dict(help='store run-process services after every chain',
                                       action='store_true'),
                        store_one=dict(help='store run-process services after chain CHAIN_NAME',
                                       metavar='CHAIN_NAME'),
                        store_none=dict(help='do not store run-process services',
                                        action='store_true'),
                        results_dir=dict(help='set directory path for results output',
                                         metavar='RESULTS_DIR'),
                        data_dir=dict(help='set directory path for data',
                                      metavar='DATA_DIR'),
                        macros_dir=dict(help='set directory path for macros',
                                        metavar='MACROS_DIR'),
                        templates_dir=dict(help='set directory path for template files',
                                           metavar='TEMPLATES_DIR'),
                        spark_cfg_file=dict(help='set path of Spark configuration file',
                                            metavar='SPARK_CONFIG_FILE'),
                        seed=dict(help='set seed for random-number generation',
                                  action='append',
                                  metavar='KEY=SEED'), )

USER_OPTS_CONF_KEYS = dict(analysis_name='analysisName',
                           analysis_version='version',
                           batch_mode='batchMode',
                           log_level='logLevel',
                           log_format='logFormat',
                           profile='doCodeProfiling',
                           begin_with='beginWithChain',
                           end_with='endWithChain',
                           store_all='storeResultsEachChain',
                           store_one='storeResultsOneChain',
                           store_none='doNotStoreResults',
                           spark_cfg_file='sparkCfgFile',
                           seed='seeds', )


def set_opt_var(opt_key, settings, args):
    """Set configuration variable from user options."""
    value = args.get(opt_key)
    if value is None:
        return
    conf_key = USER_OPTS_CONF_KEYS.get(opt_key, opt_key)
    settings[conf_key] = CONFIG_TYPES.get(conf_key, str)(value)


CONFIG_OPTS_SETTERS = collections.defaultdict(lambda: set_opt_var)


def set_log_level_opt(opt_key, settings, args):
    """Set configuration log level from user option."""
    level = args.get(opt_key)
    if not level:
        return
    if level not in LogLevel.__members__:
        raise ValueError('invalid logging level specified: "{!s}"'.format(level))

    settings[USER_OPTS_CONF_KEYS.get(opt_key, opt_key)] = level


CONFIG_OPTS_SETTERS['log_level'] = set_log_level_opt


def set_begin_end_chain_opt(opt_key, settings, args):
    """Set begin/end-chain variable from user option."""
    chain = args.get(opt_key)
    if not chain:
        return
    if args.get('single_chain'):
        raise RuntimeError('"begin-with" and "end-with" chain options cannot be combined with "single-chain" option')
    settings[USER_OPTS_CONF_KEYS.get(opt_key, opt_key)] = str(chain)


CONFIG_OPTS_SETTERS['begin_with'] = set_begin_end_chain_opt
CONFIG_OPTS_SETTERS['end_with'] = set_begin_end_chain_opt


def set_single_chain_opt(opt_key, settings, args):
    """Set single-chain variable from user option."""
    chain = args.get(opt_key)
    if not chain:
        return
    settings[USER_OPTS_CONF_KEYS['begin_with']] = str(chain)
    settings[USER_OPTS_CONF_KEYS['end_with']] = str(chain)


CONFIG_OPTS_SETTERS['single_chain'] = set_single_chain_opt


def set_seeds(opt_key, settings, args):
    """Set random seeds."""
    seed_args = args.get(opt_key)
    if not seed_args:
        return

    seeds = settings[USER_OPTS_CONF_KEYS.get(opt_key, opt_key)]
    for kv in seed_args:
        kv = kv.strip()
        eq_pos = kv.find('=')
        if eq_pos == 0 or eq_pos == len(kv) - 1:
            raise RuntimeError('expected "key=seed" for --seed command-line argument; got "{}"'.format(kv))
        key, value = (kv[:eq_pos].strip().lower(), kv[eq_pos + 1:].strip()) if eq_pos > 0 else ('default', kv.strip())
        seeds[key] = value


CONFIG_OPTS_SETTERS['seed'] = set_seeds


def set_custom_user_vars(opt_key, settings, args):
    """Set custom user configuration variables."""
    custom_vars = args.get(opt_key)
    if not custom_vars:
        return

    for var in custom_vars:
        # parse key-value pair
        var = var.strip()
        eq_pos = var.find('=')
        if eq_pos < 1 or eq_pos > len(var) - 2:
            raise RuntimeError('Expected "key=value" for --conf-var command-line argument; got "{}"'.format(var))
        key, value = var[:eq_pos].strip(), var[eq_pos + 1:].strip()

        # interpret type of value
        try:
            settings[key] = ast.literal_eval(value)
        except Exception:
            settings[key] = value


CONFIG_OPTS_SETTERS['conf_var'] = set_custom_user_vars



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk201_readdata

Created: 2017/02/17

Description:
    Macro to that illustrates how to open files as pandas datasets.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain, resources
from eskapade import core_ops, analysis
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk201_readdata')

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk201_readdata'
settings['version'] = 0

#########################################################################################
# --- Analysis configuration flags.
#     E.g. use these flags turn on or off certain chains with links.
#     by default all set to false, unless already configured in
#     configobject or vars()

# turn on/off the 2 examples
settings['do_example1'] = True
settings['do_example2'] = True

#########################################################################################
# --- Set path of data
data_path = resources.fixture('dummy.csv')

#########################################################################################
# --- now set up the chains and links, based on configuration flags

# --- example 1: readdata with one input file
if settings['do_example1']:
    ch1 = Chain('MyChain1')
    read_data = analysis.ReadToDf(key='test1', sep='|', reader='csv', path=data_path)
    ch1.add(read_data)

    # --- do something useful with the test dataset here ...

# --- example 2: readdata with default settings reads all three input files simultaneously.
#                all extra key word arguments are passed on to pandas reader.
if settings['do_example2']:
    ch2 = Chain('MyChain2')

    # --- a loop is set up in the chain MyChain.
    #     we iterate over (chunks of) the next file in the list until the iterator is done.
    #     then move on to the next chain (Overview)

    # --- readdata keeps on opening the next file in the file list.
    #     all kwargs are passed on to pandas file reader.
    read_data = analysis.ReadToDf(name='reader2', key='test2', sep='|', reader='csv', usecols=['x', 'y'])
    read_data.path = [data_path] * 3
    ch2.add(read_data)

# --- print contents of the datastore
overview = Chain('Overview')
pds = core_ops.PrintDs(name='End')
pds.keys = ['n_test1', 'n_test2']
overview.add(pds)

#########################################################################################

logger.debug('Done parsing configuration file esk201_readdata')
